import numpy as np
from abc import ABCMeta, abstractmethod
from qdarts.util_functions import find_label
from qdarts.simulator import AbstractPolytopeSimulator


def softmax(v, axis=None):
    max_v = np.max(v)
    y = np.exp(v - max_v)
    return y / np.sum(y, axis)


class AbstractSensorSim(metaclass=ABCMeta):
    """Base class defining the interface for all sensor simulations"""

    def __init__(self, num_sensors):
        """Initialized a sensor configuration with num_sensors sensor dots"""
        self.num_sensors = num_sensors

    @abstractmethod
    def slice(self, P, m):
        """Takes an affine subspace of the simulated model."""
        pass

    @abstractmethod
    def start_measurement(self):
        """Reinitializes the sensor as to generate independent noise samples"""
        pass

    @abstractmethod
    def precompute_sensor_state(self, state, A, b, basis_labels):
        """Allows the sensor to precompute internal information that is valid for a whole ground state polytope.

        This allows the sensor to precompute and cache information that is valid for all voltages v that are inside a
        ground state polytope P. The returned values are cached in the polytope objects of the simulator and
        supplied as sensor_state argument during a call of sample_sensor_equilibrium and sample_sensor_configuration.
        The supplied information provides all information of the basis labels considered by the simulation for P(n),
        and the linear functions defining the facets of P(n) Av+b. Note that as everywhere else, these linear functions define
        energy differences for each considered basis label to the ground state.

        Parameters
        ----------
        state: np.array of ints
            the state of N dots, identifying the ground state polytope for which to generate the sensor state information
        A: LxK np.array of floats
            LxK linear parameters of the energy difference function for the K sensor gates
        b: np.array of floats
            The affine offsets of the energy difference function for the L basis states
        basis_labels: LxN np.array of ints
            The labels of the L basis states
        """
        pass

    @abstractmethod
    def sample_sensor_equilibrium(self, v, H, mixed_state, sensor_state, beta):
        """Computes a noisy average of the sensor response for a given mixed state.

        This is intended to simulate a long (>1mus) time integration of the sensor signal, thus
        we can assume that states are thermalized but the signal is still affected by noise.

        Parameters
        ----------
        v: np.array of floats
            vector of K gate voltages defining the current system
        H: LxL np.array of floats
            Hamiltonian of the system defined by v. Labels and basis are the same as in precompute_sensor_state
        mixed_state: LxL np.array of floats
            Mixed state matrix computed via the Hamiltonian using expm(-beta*H)
        sensor_state:
            Cached information returned by precompute_sensor_state. All information therein are internal to the
            sensor simulator
        beta: float
            scaled inverse temperature parameter
        """
        pass

    @abstractmethod
    def sample_sensor_configuration(
        self, sampled_configuration, v, H, mixed_state, sensor_state, beta
    ):
        """samples a sensor response for a given sampled elecron configuration

        This is intended to simulate a short (<<1mus) time integration of the sensor signal,
        where we can not assume that electrons transitions during the measurement. In this case,
        the user supplied the relevant configuration and the sensor returns a sampled signal for this configuration.
        Care should be taken that the configuration sampled has all information needed in the base to compute the sensor
        signal, e.g., there should be a state with one more or less electrons on each sensor dot.


        Parameters
        ----------
        sampled_configuration: np.array of ints
            vector of N elements describing the sampled electron configuration.
        v: np.array of floats
            vector of K gate voltages defining the current system
        H: LxL np.array of floats
            Hamiltonian of the system defined by v. Labels and basis are the same as in precompute_sensor_state
        mixed_state: LxL np.array of floats
            Mixed state matrix computed via the Hamiltonian using expm(-beta*H)
        sensor_state:
            Cached information returned by precompute_sensor_state. All information therein are internal to the
            sensor simulator
        beta: float
            scaled inverse temperature parameter
        """
        pass


class NoisySensorDot(AbstractSensorSim):
    """Simulates a sensor signal by computing the conductance of the sensor dots.

    This class implements the interface of AbstractSensorSim and for most points, the
    documentation there should be referred to. This simulation combines a simple
    estimation of the conductance g with two noise sources. A fast noise source
    that simulates gaussian white noise that is drawn for each query of the sensor response
    and a slow noise source that models time dependent noise between different invocations
    of the sensor.

    The shape of the simulated sensor peak can be configured ´via config_peak, in which
    height and with of the peak can be adapted. Currently, all sensor dots share these
    parameters.

    The noise can be configured via config_noise. The noise is modeled as an additive
    noise on the sensor peak position in voltage space. Thus, at peaks or valleys, the noise
    is small while on the sides of the peak, where the derivatives are largest, the noise will
    affect measurements the most. Additional signal  noise is then modeled by adding white gaussian noise.
    """

    def __init__(self, sensor_dot_ids):
        super().__init__(len(sensor_dot_ids))
        self.sensor_dot_ids = sensor_dot_ids
        self.g_max = 1
        self.fast_noise_var = 0.0
        self.peak_width_multiplier = 1
        self.slow_noise_gen = None
        self.signal_noise_scale = 0.0

    def config_noise(self, sigma, signal_noise_scale, slow_noise_gen=None):
        self.fast_noise_var = sigma**2
        self.slow_noise_gen = slow_noise_gen
        self.signal_noise_scale = signal_noise_scale
        # initialize noise
        self.start_measurement()

    def config_peak(self, g_max, peak_width_multiplier):
        self.g_max = g_max
        self.peak_width_multiplier = peak_width_multiplier

    def start_measurement(self):
        if self.slow_noise_gen is not None:
            self.slow_noise_gen.start_sequence()

    def slice(self, P, m):
        # if there is no slow noise, there is nothing to slice
        if self.slow_noise_gen is None:
            return self

        # otherwise create a copy of this with a sliced slow noise model
        sliced_sensor_dot = NoisySensorDot(self.sensor_dot_ids)
        sliced_sensor_dot.g_max = self.g_max
        sliced_sensor_dot.fast_noise_var = self.fast_noise_var
        sliced_sensor_dot.peak_width_multiplier = self.peak_width_multiplier
        sliced_sensor_dot.signal_noise_scale = self.signal_noise_scale
        sliced_sensor_dot.slow_noise_gen = self.slow_noise_gen.slice(P, m)
        return sliced_sensor_dot

    def precompute_sensor_state(self, state, A, b, labels):
        sensor_state = {}
        for i, sensor_id in enumerate(self.sensor_dot_ids):
            labels_nosens = np.delete(labels, sensor_id, axis=1)
            labels_unique, inverse_index = np.unique(
                labels_nosens, return_inverse=True, axis=0
            )

            labels_sens = labels[:, sensor_id]
            sorted_ind = np.lexsort((labels_sens, inverse_index))

            relevant_label_indices = []
            prev = []
            nex = []
            cur = -1
            last = None
            last_2 = None
            for ind in sorted_ind:
                lab = labels_nosens[ind]
                if np.any(lab != cur):
                    cur = lab
                    last = None
                    last_2 = None
                else:
                    if last_2 is not None:
                        relevant_label_indices.append(last)
                        prev.append(last_2)
                        nex.append(ind)
                    last_2 = last
                    last = ind
            terms = np.array(relevant_label_indices, dtype=int)
            prev = np.array(prev, dtype=int)
            nex = np.array(nex, dtype=int)
            terms_labels = labels[terms, :]
            sensor_state[sensor_id] = (terms, prev, nex, terms_labels)
        return sensor_state

    def _precompute_g(self, v, H, sensor_state, beta):
        results = np.zeros(len(self.sensor_dot_ids))
        gs = {}
        slow_noise = np.zeros((results.shape[0], 1))
        if self.slow_noise_gen is not None:
            slow_noise = self.slow_noise_gen(v)
        for i, sensor_id in enumerate(self.sensor_dot_ids):
            terms, neighbour_prev, neighbour_next, _ = sensor_state[sensor_id]

            # compute sensor detuning between every state and their neighbour
            eps_prev = np.abs(np.diag(H)[terms] - np.diag(H)[neighbour_prev])
            eps_next = np.abs(np.diag(H)[terms] - np.diag(H)[neighbour_next])
            eps = np.minimum(eps_prev, eps_next)
            # add noise
            eps = eps + slow_noise[i : i + 1]
            if self.fast_noise_var > 0:
                fast_noise = np.random.randn(*eps.shape) * np.sqrt(self.fast_noise_var)
                eps += fast_noise
            eps *= beta

            # we approximate the logistic peak of g with the peak of a normal distribution of same width
            # todo: we can fully go back to the logistic peak
            var_logistic = (1 / 0.631 * self.peak_width_multiplier) ** 2

            def norm_pdf(x, mu, var):
                return (
                    1 / np.sqrt(2 * np.pi * var) * np.exp(-((x - mu) ** 2) / (2 * var))
                )

            gs[sensor_id] = self.g_max * 4 * norm_pdf(0, eps, var_logistic)
        return gs

    def sample_sensor_equilibrium(self, v, H, mixed_state, sensor_state, beta):
        results = np.zeros(len(self.sensor_dot_ids))
        gs = self._precompute_g(v, H, sensor_state, beta)

        for i, sensor_id in enumerate(self.sensor_dot_ids):
            terms, neighbour_prev, neighbour_next, _ = sensor_state[sensor_id]
            g = gs[sensor_id]
            p = np.diag(mixed_state)[terms]
            results[i] = np.sum(p * g) / np.sum(p)
        var_logistic = (1 / 0.631 * self.peak_width_multiplier) ** 2
        scale = (
            self.g_max * self.signal_noise_scale * 4 / np.sqrt(2 * np.pi * var_logistic)
        )
        results += scale * np.random.randn(len(results))
        return results

    def sample_sensor_configuration(
        self, sampled_configuration, v, H, mixed_state, sensor_state, beta
    ):
        results = np.zeros(len(self.sensor_dot_ids))
        gs = self._precompute_g(v, H, sensor_state, beta)

        for i, sensor_id in enumerate(self.sensor_dot_ids):
            terms, neighbour_prev, neighbour_next, terms_labels = sensor_state[
                sensor_id
            ]
            label_pos = find_label(terms_labels, sampled_configuration)
            results[i] = gs[sensor_id][label_pos]
        var_logistic = (1 / 0.631 * self.peak_width_multiplier) ** 2
        scale = (
            self.g_max * self.signal_noise_scale * 4 / np.sqrt(2 * np.pi * var_logistic)
        )
        results += scale * np.random.randn(len(results))
        return results


class TunnelBarrierModel:
    """Model of the tunnel barriers of a device

    This class defines a mapping between gate voltages of the device and
    the tunnel coupling between the dots. To be more exact, the tunnel
    coupling between dots i, and j and the gate voltages v is given by

    T_ij = exp(W_ij^Tv+b_ij)

    where W_ij is a vector of couplings and b_ij is an offset.
    """

    def __init__(self, gate_offsets, gate_levers=None):
        """Creates a tunnel barrier model.

        Parameters
        ----------
        gate_offsets : NxN np.array of floats
            the offsets b_ij provided as matrix.
        gate_levers : NxNxK np.array of floats or None
            Here, K is the number of plunger gates. The first two indices describe the index of the tunnel coupling matrix ij.
            If None, it is assumed to be 0.
        """
        self.gate_offsets = gate_offsets
        self.gate_levers = gate_levers

    def slice(self, P, m):
        """Takes an affine subspace of the simulated model.

        Let v=Pv'+m. Computes a new parameterization such, that
        T_ij = exp(W'_ij^Tv'+b'_ij)

        Parameters
        ----------
        P : KxM np.array of floats
            The linear transformation matrix
        m : np.array of floats or None
            Array of size K storing the affine offset.
        """
        if self.gate_levers is None:
            return self
        sliced_levers = self.gate_levers.reshape(-1, self.gate_levers.shape[-1]) @ P

        sliced_levers = sliced_levers.reshape(
            self.gate_levers.shape[0], self.gate_levers.shape[1], P.shape[1]
        )
        m_applied = self.gate_levers.reshape(-1, self.gate_levers.shape[-1]) @ m
        sliced_offsets = self.gate_offsets + m_applied.reshape(
            self.gate_levers.shape[0], self.gate_levers.shape[1]
        )
        return TunnelBarrierModel(sliced_offsets, sliced_levers)

    def get_tunnel_matrix(self, v):
        """Returns the tunnel matrix for a given gate voltage

        Parameters
        ----------
        v : np.array of floats or None
            Array of size K storing the gate voltages

        """
        if self.gate_levers is None:
            return np.exp(self.gate_offsets)
        else:
            barrier_potentials = self.gate_levers @ v
            return np.exp(barrier_potentials + self.gate_offsets)


class LocalSystem:
    """Class describing a quantum system defined by the gate voltages of a simulated device.

    For a given set of gate voltages, the simulator first computes a set of core states that are most
    likely relevant for the computation of the hamiltonian and then extends it by adding additional
    states. These are then used to define a basis of the vector space for the Hamiltonian, which
    is then used to compute the mixed state. Finally, the mixed state is then used to simulate a sensor signal.

    This class stores all this information and includes some minimal tools to query information on
    different sub-bases. This class is tightly coupled to tunneling_simulator.

    Attributes
    ----------

    v: np.array of floats
        gate voltages that define the parameters of this system
    state: np.array of ints
        the ground state configuration of v
    beta: float
        the scaled inverse temperature 1/k_bT
    H: LxL np.array of floats
        Hamiltonian over the subspace spanned by the L basis state of the extended basis.
        See methods basis_labels and core_basis_indices
    """

    def __init__(self, v, H, state, sim):
        """Creates the LocalSystem.

        This is an internal function used by the tunneling simulator.
        """
        self.v = v
        self.H = H
        self.state = state.copy()
        self._sim = sim
        self.beta = self._sim.beta

    def _compute_mixed_state(self, H):
        diffs = np.diag(H) - np.min(np.diag(H))
        sel = np.where(diffs < 2 * self._sim.poly_sim.get_maximum_polytope_slack())[0]
        H_sel = H[:, sel][sel, :]

        eigs, U = np.linalg.eigh(H_sel)
        ps = softmax(-eigs * self.beta)
        rho_sel = U @ np.diag(ps) @ U.T

        rho = np.zeros(H.shape)
        indizes = H.shape[0] * sel[:, None] + sel[None, :]
        np.put(rho, indizes.flatten(), rho_sel.flatten())
        return rho

    @property
    def mixed_state(self):
        """Computes an approximate mixed state matrix over the full basis.

        the mixed state matrix, defined as expm(-beta*H)

        Note that this function approximated the true mixed state matrix by inly taking basis eleemnts into account
        that have a small energy difference to the ground state. This is a multiple of the polytope slack used by the
        capacitive simulation.
        """
        return self._compute_mixed_state(self.H)

    def compute_mixed_state_of_subset(self, subset_indices):
        """Computes the mixed state for a subset, ignoring the existance of any other state entirely.

        The result is a KxK matrix where K is the length of subset_indices.
        This function is not equivalent to selecting a subset of mixed_state, since this assumes that
        the states not referenced by subset_indices are ruled out for some other reason, i.e., they are
        assigned probability 0 and probabilities are renormalized to sum to 1 over the elements in the subset.


        Parameters
        ----------
        subset_indices : np.array of ints
            The L' indices into the basis element matrix as returned by basis_labels
        m : np.array of floats or None
        """

        return self._compute_mixed_state(self.H[subset_indices, :][:, subset_indices])

    @property
    def basis_labels(self):
        """The labels of the basis elements, indentified by their ground state electron configuration"""
        return (
            self._sim.boundaries(self.state).additional_info["extended_polytope"].labels
        )

    @property
    def core_basis_indices(self):
        """Indices into basis_labels that define the subset of core basis elements."""
        return (
            self._sim.boundaries(self.state)
            .additional_info["extended_polytope"]
            .core_basis_indices
        )

    def sample_sensor_equilibrium(self):
        """Samples a boisy averaged sensor response from the current system over all basis elements.

        This returns the average signal with added sensor noise. This is an approximation to long
        average measurements at a single point.
        """
        sensor_state = self._sim.boundaries(self.state).additional_info["sensor_state"]
        return self._sim.sensor_sim.sample_sensor_equilibrium(
            self.v, self.H, self.mixed_state, sensor_state, self.beta
        )

    def sample_sensor_configuration(self, sampled_configuration):
        """Samples the sensor signal for a given sampled electron configuration.

        For a short time simulation it is more prudent to externally sample a state from the basis and then generate a sensor signal
        from it. This function allows this. Note that only selecting states from the set of core_basis_indices is safe as otherwise
        the sensor might miss information required to correctly compute the response.

        Parameters
        ----------
            sampled_configuration: list of ints
                the sampled state for which to generate the sensor response.
        """
        sensor_state = self._sim.boundaries(self.state).additional_info["sensor_state"]
        return self._sim.sensor_sim.sample_sensor_configuration(
            sampled_configuration,
            self.v,
            self.H,
            self.mixed_state,
            sensor_state,
            self.beta,
        )


class ApproximateTunnelingSimulator(AbstractPolytopeSimulator):
    """Simulator for approximate charge tunneling in a quantum dot device.

    The simulator extends the supplied capacitive simulation by creating a Hamiltionian H,
    where on the diagonals are the capacitive energies of the simualation, while the off-diagonals
    have added tunnel coupling parameters. Locally the hamiltonian is approximated via L basis states,
    where each state is an electron configurtion on the dots. This mixed state is then used to create a sensor simulation.

    It is possible to query the state of single hamiltonian, their mixed state and their sensor simulation via the class
    LocalSystem, returned by compute_local_system, but the primary use of tis class lies in its ability to compute
    1D or 2D sensor scans via sensor_scan and sensor_scan_2D.

    For computing the tunnel coupling parameters, this class can make use of an additional Tunnel barrier simulation, but it
    is also possible to just supply a NxN constant matrix of tunnel couplings between all D dots in the array.

    Finally, the class follows the interface of AbstractPolytopeSimulator, which means it is possible to directly query the information
    of the underlying polytopes of the simulation. This is there to unify slicing between simulators.

    Implementation details:

    The basis used for a gate voltage v is queried by finding its ground states n and then the facets of the
    ground state polytope P(n) create the basis. Thus, this basis becomes extended as the slack variable in the underlying
    capacitance simulation is increased. This is called the core state set.
    Additionally, the simulation allows to add additional states. For example, for most sensor simulations to work, we also need
    other higher energy states to compute correct conductance. These additional states can be added by modifying the vector
    num_additional_neighbours. if the ith element in this vector is R>0, and s is a state in the core basis, then
    the extended basis will also include the states :math:`s+ke_i` where :math:`|k|<=R` and :math:`e_i` is the ith basis vector.

    The tunnel couplings T are included into the Hamiltonian the following way: let :math:`s_i` and :math:`s_j` be two states in the basis of the Hamiltonian
    that differ only in the value of the electron configuration at dots i and j.
    More exactly, we have that :math:`s_i` and :math:`s_j` are related by moving an electron from state :math:`s_i` to :math:`s_j` or vice versa.
    Let :math:`H_{kl}` be the off-diagonal matrix element of those states. Then we have :math:`H_{kl} = T_{ij}`.
    In all other cases, tunnel coupling is 0.

    The mixed state is then again computed approximately, for more info on that, see documentation of LocalSystem.

    The sensor signal of the computed mixed state is computed via the sensor_sim.

    Attributes
    ----------

    beta: float
        Scaled inverse temperature 1/k_BT
    T: float
        Temperature
    poly_sim:
        the capacitive simulation object
    barrier_sim:
        the barrier simulation object. Note that even if the supplied object to init was a matrix, this will be a TunnelBarrierModel.
    sensor_sim:
        the sensor simulation object
    num_additional_neighbours: np.array of ints
        for each dot defines how many additional states should be added for each state in the core basis. This is done
        by adding or subtracting electrons on the ith element where the maximum is given by the ith element of num_additional_neighbours.
        We advise to set this to 2 for sensor dots. Note that computation time can quickly explode when increasing this parameter.
        Outside of sensor dots, we advise therefore to increase the slack in the capacitive simulation.
    """

    def __init__(self, polytope_sim, barrier_sim, T, sensor_sim):
        """Creates a tunneling simulation

        Parameters
        ----------
        polytope_sim:
            capacitance simulator object that computes ground state polytopes and capacitive energy differences
        barrier_sim: Object or Matrix
            Either a DxD basis that describes a constant tunnel coupling between all D dots. Note that the diagonal of this matrix is zero.
            Alternatively an object with a method barrier_sim.get_tunnel_matrix(v) returning a DxD matrix, and which supports the slice operation.
        T: float
            Temperature in Kelvin. Good values are < 0.1
        sensor_sim: Derived from AbstractSensorSim
            A sensor simulation that follows the interface of AbstractSensorSim and which computes the sensor signal.
        """
        self.poly_sim = polytope_sim
        # compatibility to earlier code that uses matrices
        if isinstance(barrier_sim, np.ndarray):
            self.barrier_sim = TunnelBarrierModel(np.log(barrier_sim + 1.0e-20))
        else:
            self.barrier_sim = barrier_sim
        eV = 1.602e-19
        kB = 1.380649e-23 / eV
        self.beta = 1.0 / (kB * T)
        self.T = T
        self.sensor_sim = sensor_sim

        # clean up potentially stored conflicting data
        for poly in self.poly_sim.cached_polytopes():
            poly.additional_info.pop("features_out_info", None)

        self.num_additional_neighbours = np.zeros(self.poly_sim.num_dots, dtype=int)

        super().__init__(self.poly_sim.num_dots, self.poly_sim.num_inputs)

    def slice(self, P, m, proxy=False):
        """Restricts the simulator to the affine subspace v=m+Pv'

        Computes the slice through the simulated device by setting v=m+Pv', where v is the plunger gate voltages of the
        original device and v' is the new coordinate system. This is implemented here by slicing all the different parts
        of the simulation, capacitance model, barrier model and sensor model.

        Parameters
        ----------
        P : MxK np.array of floats
            The linear coefficient matrix. Here M is the number of voltage elements in v in the full simulation
            and K the dimensionality of the subspace.
        m: offset of the affine trnsformation.
        proxy: bool
            Whether a proxy is returned. A proxy shares the cache, if possible. This is the case when P is invertible,
            especially this entails M=K. If cache sharing is possible, the simulation computes the original polytope and then
            applies the affine transformation. This can reduce run time a lot if several slices need to be computed for the
            same simulation.

        Returns
        -------
        A simulator object describing the simulation on the affine subspace. The current simulation object remains unchanged.
        """
        sliced_poly_sim = self.poly_sim.slice(P, m, proxy)
        sliced_barrier_sim = self.barrier_sim.slice(P, m)
        sliced_sensor_sim = self.sensor_sim.slice(P, m)
        sliced_tunneling_sim = ApproximateTunnelingSimulator(
            sliced_poly_sim, sliced_barrier_sim, self.T, sliced_sensor_sim
        )
        sliced_tunneling_sim.num_additional_neighbours = (
            self.num_additional_neighbours.copy()
        )
        return sliced_tunneling_sim

    def _compute_tunneling_op(self, state_list):
        """Computes the mapping between tunnel coupling and hamiltonian off diagonal elements
        and also also computes a multiplicative weight for each tunnel coupling based on the number
        of affected electrons during the state transitions.

        We currently only add tunnel coupling between two states n,m if they describe the transition of a single
        electron between two dots i and j. In this case the tunnel coupling is w*T_ij where w=1 if the
        total number of electrons on dots i and j is odd, otherwise w=sqrt(2).
        Parameters
        ----------
        state_list: list of vectors of ints
            The list of states that describe a subset of the fokh basis of the Hamiltonian.

        Returns
        -------
        TOp: a mapping TOP(n,m)=i*num_dots+j, the index in the flattened matrix of tunnel couplings
        TOpW: weight matrix W(n,m).

        """
        N = state_list.shape[0]
        n_dots = state_list.shape[1]
        TOp = np.zeros((N, N), dtype=int)
        TOpW = np.ones((N, N))

        sums = np.sum(state_list, axis=1)
        for i, s1 in enumerate(state_list):
            for j, s2 in zip(range(i + 1, len(state_list)), state_list[i + 1 :]):
                if sums[i] != sums[j]:
                    continue

                if np.sum(np.abs(s1 - s2)) == 0:
                    continue
                abs_diff = np.abs(s1 - s2)
                if np.sum(abs_diff) != 2:
                    continue

                # compute lookup indices in tunneling strength matrix
                idxs = np.where(abs_diff > 0)[0]
                if len(idxs) == 1:
                    ind = idxs[0] * n_dots + idxs[0]
                else:
                    ind = idxs[0] * n_dots + idxs[1]
                TOp[i, j] = ind
                TOp[j, i] = ind

                # compute weight. If the total number of electrons on the affected dots
                # is even, the tunneling strength is multiplied by sqrt(2)
                pos_changes = s1 != s2
                num_electrons_affected = np.sum(pos_changes * s1)
                if num_electrons_affected % 2 == 0:
                    TOpW[i, j] = np.sqrt(2)
                    TOpW[j, i] = TOpW[i, j]
        return TOp, TOpW

    def _create_state_list(self, state, direct_neighbours):
        """Creates the extended basis"""
        state_list = np.vstack([direct_neighbours, [np.zeros(len(state), dtype=int)]])

        additional_states = []
        for i in range(self.poly_sim.num_dots):
            e_i = np.eye(1, self.poly_sim.num_dots, i, dtype=int)
            for k in range(1, 1 + self.num_additional_neighbours[i]):
                additional_states.append(state_list + k * e_i)
                additional_states.append(state_list - k * e_i)

        if len(additional_states) > 0:
            for add in additional_states:
                state_list = np.vstack([state_list, add])
            state_list = np.unique(state_list, axis=0)
        state_list += state[None, :]
        state_list = state_list[np.all(state_list >= 0, axis=1)]

        # mark the subset of original polytope transitions + the current state

        core_index_set = [int(find_label(state_list, state)[0])]
        for core_transition in direct_neighbours:
            core_index_set.append(
                int(find_label(state_list, core_transition + state)[0])
            )
        return state_list, np.array(core_index_set, dtype=int)

    def boundaries(self, state):
        """
        Returns the polytope P(n) of a given state n with all its boundaries, labels and meta information.

        If the polytope is not cached, it needs to be computed. This can take some time for large devices.

        Parameters
        ----------
        state: list of ints
            The state n for which to compute the polytope P(n)

        Returns
        -------
        The polytope P(n)
        """
        state = np.asarray(state)
        polytope = self.poly_sim.boundaries(state)
        # cache features_out info in polytope structure
        if "extended_polytope" not in polytope.additional_info.keys():
            # create a list of all neighbour states of interest for use in the Hamiltonian
            state_list, polytope_base_indx = self._create_state_list(
                state, polytope.labels
            )

            # create full set of transition equations
            A, b = self.poly_sim.compute_transition_equations(state_list, state)

            TOp, TOpW = self._compute_tunneling_op(state_list)
            extended_polytope = type("", (object,), {})()
            extended_polytope.A = A
            extended_polytope.b = b
            extended_polytope.TOp = TOp
            extended_polytope.TOpW = TOpW
            extended_polytope.labels = state_list
            extended_polytope.core_basis_indices = polytope_base_indx
            polytope.additional_info["extended_polytope"] = extended_polytope

            # also compute the sensor info
            polytope.additional_info["sensor_state"] = (
                self.sensor_sim.precompute_sensor_state(state, A, b, state_list)
            )
        return polytope

    def _create_hamiltonian(self, v, A, b, TOp, TOpW):
        """Computes the hamiltonian at the given gate voltages"""
        tunnel_matrix = self.barrier_sim.get_tunnel_matrix(v)
        N = A.shape[0]
        energy_diff = -(A @ v + b)
        if tunnel_matrix is None:
            return np.diag(energy_diff)
        else:
            t_term = ((tunnel_matrix.reshape(-1)[TOp.reshape(-1)]).reshape(N, N)) * TOpW
            return np.diag(energy_diff) - t_term

    def get_displacement(self, H, dH):
        """Computes the displacement of the ground state"""
        ind0 = np.argsort(np.diag(H))[:2]
        tc = H[ind0[0], ind0[1]]
        eps = H[ind0[0], ind0[0]] - H[ind0[1], ind0[1]]

        dind0 = np.argsort(np.diag(dH))[:2]
        dtc = dH[dind0[0], dind0[1]]
        deps = dH[dind0[0], dind0[0]] - dH[dind0[1], dind0[1]]

        return deps / np.sqrt(dtc**2 + deps**2) - eps / np.sqrt(tc**2 + eps**2)

    def compute_local_system(self, v, state, search_ground_state=True):
        """Computes a full description of the local quantum system and returns the LocalSystem object.

        This is a full locla simulation of the device and can be used to query sensor values but also the mixed state matrix.
        See LocalSystem for more info.

        Note that unlike in most other places, v does not need to belong to the ground state polytope of state.
        This might be useful for the computation of signals in which the device is far out of equilibrium.

        Parameters
        ----------
        v: np.array of floats
            The vector of gate voltages of the device
        state: np.array of ints
            The ground state polytope relative to which the local system is computed. This is in most cases the ground state.
        search_ground_state: bool
            If True, verifies that state is the ground state of v and searches it otherwise. If you know that this is the case,
            you can safely set it to False for a speed-up. In the general case, setting this to false will compute the
            LocalSystem relative to a different basis state.
        """
        if search_ground_state:
            state = self.poly_sim.find_state_of_voltage(v, state_hint=state)
        polytope = self.boundaries(state)
        extended_polytope = polytope.additional_info["extended_polytope"]
        H = self._create_hamiltonian(
            v,
            extended_polytope.A,
            extended_polytope.b,
            extended_polytope.TOp,
            extended_polytope.TOpW,
        )
        system = LocalSystem(v, H, state, self)
        return system

    def sensor_scan(
        self,
        v_start,
        v_end,
        resolution,
        v_start_state_hint,
        cache=True,
        start_new_measurement=True,
        insitu_axis=None,
    ):
        """Computes a 1D sensor ramp scan.

        Computes a linear set of points between v_start and v_end and for each point computes the sensor signal.
        To be more exact, for each point, the ground state polytope is computed which is then used to define the local_system.
        Returns the sensor signal for each sensor and dot

        Parameters
        ----------
        v_start: np.array of floats
             Vector of gate voltages of the device describing the first measurement point
        v_end: np.array of floats
             Vector of gate voltages of the device describing the last measurement point
        resolution: int
            number of measured points on the linear scan between v_start and v_end, including both end points.
        v_start_state_hint: np.array of int
            Guess for the state n for which holds that v_start is element of P(n). The simulator will use this
            guess as a starting point for the search of the correct state if this guess is wrong. Note that P(n)
            must intersect with the affine slice, if slicing was used.
        cache: bool
            Whether the simulation should try to cache the computed polytopes. This might lead to a slower computation time
            for a scan compared to not using caching, but consecutive scans with similar ranges tend to be quicker.
        start_new_measurement: bool
            Whether the seimulated sensor measurement should be independent of any previous measurements.
        insitu_axis: list or None
            The axis along which the modulated signal is applied, insitu_axis @ plane_axes. If none, the code uses standard sensor dot approach. If a list it computes changes in quantum capacitance
        """
        # prepare start state
        state = self.poly_sim.find_state_of_voltage(
            v_start, state_hint=v_start_state_hint
        )

        P = (v_end - v_start).reshape(-1, 1)
        if cache:
            sim_slice = self.slice(P, v_start, proxy=cache)
        else:
            sim_slice = self

        if start_new_measurement:
            sim_slice.sensor_sim.start_measurement()

        if insitu_axis is None:
            values = np.zeros((resolution, self.sensor_sim.num_sensors))
        else:
            values = np.zeros((resolution, 1))
        for i, v0 in enumerate(np.linspace([0.0], [1.0], resolution)):
            if cache:
                v = v0
            else:
                v = v_start + P @ v0
            if not sim_slice.poly_sim.inside_state(v, state):
                state = sim_slice.poly_sim.find_state_of_voltage(v, state_hint=state)

            system = sim_slice.compute_local_system(v, state, search_ground_state=False)
            if insitu_axis is None:
                values[i] = system.sample_sensor_equilibrium()
            else:
                # inisitu reflecometry
                dv = 0.0001
                system2 = sim_slice.compute_local_system(
                    np.array(v) + dv * np.array(insitu_axis),
                    state,
                    search_ground_state=False,
                )
                a = sim_slice.get_displacement(system.H, system2.H)
                values[i] = a
        return values

    def sensor_scan_2D(
        self,
        P,
        m,
        minV,
        maxV,
        resolution,
        state_hint_lower_left,
        cache=True,
        insitu_axis=None,
    ):
        """Computes the sensor signal on a 2D grid of points.

        For the exact computation of points, see sensor_scan.

        The grid is defined the following way: Let w_ij be a 2D vector that is part of a regular
        rectangular grid spanned by the lower left corner given by minV and the upper right corner given
        by maxV and let (m,n) be the number of points in both grid directions. We have that w_00=minV and w_m-1,n-1=maxV.

        This grid is then affinely transformed into the K-dimensional space of gate vectors via
        v_ij = m+ P w_ij

        and thus P must be a Kx2 matrix and m a K-vector.

        Parameters
        ----------
        P: Kx2 np.array of floats
            linear transformation of grid-points into the K-dimensional voltage space
        m: np.array of floats:
            affine offset of the grid
        minV: np.array of floats
            2D vector describing the minimum value of the grid points
        maxV: np.array of floats
            2D vector describing the maximum value of the grid points
        resolution: int or list of ints
            if integer, describes the same number of points in both grid directions. If a list of 2 elements,
            describes the number of points along each axes of the grid.
        state_hint_lower_left: np.array of int
            Guess for the state n for point described by the grid position minV. The simulator will use this
            guess as a starting point for the search of the correct state if this guess is wrong. Note that P(n)
            must intersect with the affine slice, if slicing was used.
        cache: bool
            Whether the simulation should try to cache the computed polytopes. This might lead to a slower computation time
            for a scan compared to not using caching, but consecutive scans with similar ranges tend to be quicker.
        insitu_axis: list or None
            The axis along which the modulated signal is applied. If none, the code uses standard sensor dot approach. If a list it computes changes in quantum capacitance.
        """
        if P.shape[1] != 2:
            raise ValueError("P must have two columns")
        if isinstance(resolution, int):
            resolution = [resolution, resolution]

        # obtain initial guess for state
        line_start = self.poly_sim.find_state_of_voltage(
            m + P @ minV, state_hint=state_hint_lower_left
        )
        # now slice down to 2D for efficiency
        sim_slice = self.slice(P, m, proxy=cache)
        if insitu_axis is None:
            sim_slice.sensor_sim.start_measurement()
            values = np.zeros(
                (resolution[0], resolution[1], self.sensor_sim.num_sensors)
            )

        else:
            values = np.zeros((resolution[0], resolution[1], 1))

        for i, v2 in enumerate(np.linspace(minV[1], maxV[1], resolution[1])):
            v_start = np.array([minV[0], v2])
            v_end = np.array([maxV[0], v2])
            line_start = sim_slice.poly_sim.find_state_of_voltage(
                v_start, state_hint=line_start
            )
            values[:,i] = sim_slice.sensor_scan(
                v_start,
                v_end,
                resolution[0],
                line_start,
                cache=False,
                start_new_measurement=False,
                insitu_axis=insitu_axis,
            )
        return values



import numpy as np


class Polytope:
    r"""Represents the polytope P(n) defined by all gate voltages v in a device that have
    capacitive ground state n. They are stored as a set of linear inequalities (A,b), and
    A point v in P(n) fulfills

    :math:`Av+b <0`

    Each inequality represents a facet of the polytope and each facet marks a transition from P(n) to
    some other ground state polytope P(n'). The state difference t=n'-n is stored as label for each
    inequality.

    Not each inequality stored must touch the polytope. There might be others that could be removed without changing P(n).
    The distance from the polytope is given by the slack variable s that for each inequality either is 0 if the side is touching
    (or some number numerically close to 0, e..g, 1.e-8) and otherwise we have

    :math:`s_i = min_v A_i^Tv +b_i, v \in P(n)`

    If the inequalities measure difference in capacitive energy from the ground state (which is default in the simulator) the slack
    therefore indicates the minimum energy gap between the transition state indicated by the inequality and the ground state.

    Finally, for optimization reason, not every polytope might be fully computed and must be verified. This should never happen to a user
    and is mostly an internal detail of the simulator. This holds as well for additional information that can be stored inside a dict in the
    polytope. The simulators can store additional info in the polytope via this way.

    Attributes
    ----------
        state: D np.array of int
            The D dimensional array that stores the electron configuration n of the current ground state polytope. All points inside
            the polytope have this state as ground state.
        labels: NxD np.array of int
            Matrix of transitions. Each of the N inequalities is labeled by the state difference t=n'-n.
        A: NxK np.array of float
            Linear factors of the N inequalities for a K-dimensional gate space.
        b: N np.array of float
            constant offsets of the N linear inequalities
        slacks: N np.array of float
            Measures the distance of the inequality from the boundary of the polytope. ~0 if touching.
        point_inside: K np.array of float
            A point inside the polytope.
        must_verify: bool
            Internal variable for bookkeeping whether the polytope needs recomputing (mostly after slice). Should ALWAYS be False for polytopes
            queried from the simulator. TODO: can we remove this?
        additional_info: dict
            Internal additional information that later steps of the simulation can store inside a polytope for bookkeeping.
    """

    def __init__(self, state):
        # empty polytope
        self.state = state
        self.labels = np.array([])
        self.A = np.array([])
        self.b = np.array([])
        self.slacks = np.array([])
        self.point_inside = np.array([])
        self.must_verify = False
        self.additional_info = {}

    def set_polytope(self, labels, A, b, slacks, point_inside, must_verify=False):
        """Sets the internal variables of the polytope.

        Helper function to ensure thateverything is set as it should be.
        """
        self.labels = labels
        self.A = A
        self.b = b
        self.slacks = slacks
        self.point_inside = point_inside
        self.must_verify = must_verify

    def lazy_slice(self, P, m):
        """
        Slices a polytope lazyily, i.e., without recomputing the slacks and boundaries.

        As a result, after this must_verify is True. P is not required to be invertible.

        Parameters
        ----------
        P: KxK np.array of float
            Invertible linear transformation matrix
        m: K np.array of float
            Affine offset of the transformation.
        """
        sliced = Polytope(self.state)
        if self.A.shape[0] == 0:
            sliced.set_polytope(self.labels, self.A, self.b, np.array([]), None, False)
        else:
            sliced.set_polytope(
                self.labels,
                self.A @ P,  # we know the line equations
                self.b + self.A @ m,  # and their offsets
                None,
                None,  # but nothing else
                True,  # user must verify this later.
            )
        return sliced

    def invertible_transform(self, P, m):
        """
        Apply an invertible affine transformation to the polytope. This can be done without changing slacks and thus no verification is needed.

        Changes the space of the polytope via the transformation :math:`v=Av'+b`. Returns the polytope in the coordinate system of v'

        Parameters
        ----------
        P: KxK np.array of float
            Invertible linear transformation matrix
        m: K np.array of float
            Affine offset of the transformation.
        """
        if self.must_verify:
            return self.lazy_slice(P, m)

        transformed = Polytope(self.state)

        transformed_point_inside = np.linalg.inv(P) @ (self.point_inside - m)
        transformed.set_polytope(
            self.labels,
            self.A @ P,  # we know the line equations
            self.b + self.A @ m,  # and their offsets
            self.slacks,  # slacks are constant under invertible transforms.
            transformed_point_inside,
        )
        return transformed



import numpy as np
from qdarts.util_functions import (
    compute_maximum_inscribed_circle,
    compute_polytope_slacks,
)
from qdarts.polytope import Polytope
from abc import ABCMeta, abstractmethod


def is_sequence(seq):
    if isinstance(seq, str):
        return False
    try:
        len(seq)
    except Exception:
        return False
    return True


# internal unit conversion of capacitances from attoFarrad to Farrad/eV
eV = 1.602e-19
to_Farrad_per_eV = 1e-18 / eV


class AbstractCapacitanceModel(metaclass=ABCMeta):
    """Base Class for all capacitance models.

    This class provides all required meta information to compute capacitive energies of a system with electron configuration
    n and gate voltages v, E(v,n). The only required to this model is that E(v,n)-E(v,n') is a linear function in v and that energies
    are measured in eV.
    The class provides basic abilities: enumeration of possible transition states from a given state and computation/verification of a polytope P(n) for a state.
    For this, the user only needs to provide functions to generate transition equations and a function that allows to slice the voltage space.

    As a base class it only needs to know the number of gates/inputs and the number of dots of the array. Additionally
    the user must supply voltage bounds that ensure that all computed polytopes are bounded. In practical devices these would be voltage limits
    e.g., for device protection.

    Attributes
    ----------
    num_dots: int
            the number of discrete dot locations on the device. This is the number of elements in the electron state n.
    num_inputs: int
        The number of gate voltages of the device.
    bounds_limits: N np.array of float
        right hand side of the bound inequalities.
        set of linear inequalities A that provide bounds for the voltage space.
    """

    def __init__(self, num_dots, num_inputs, bounds_limits, bounds_normals):
        """Initializes the model.

        Parameters
        ----------
        num_dots: int
            the number of discrete dot locations on the device. This is the number of elements in the electron state n.
        num_inputs: int
            The number of gate voltages of the device.
        bounds_limits: N np.array of float
            right hand side of the bound inequalities. If bounds_normals is none, this is interpreted as lower bounds and thus N=num_inputs
        bounds_normals: N x num_inputs np.array of float
            set of linear inequalities A that provide bounds for the voltage space. A valid voltage v fulfils Av+b<0, where b is
            bounds_limits. Can be none, in which case A=-Id.
        """
        self.num_dots = num_dots
        self.num_inputs = num_inputs

        if not is_sequence(bounds_limits):
            bounds_limits = bounds_limits * np.ones(num_inputs)

        if bounds_normals is None:
            if num_inputs != len(bounds_limits):
                raise ValueError(
                    "if bounds_normals is not given, bounds_limits must be either a scalar or a sequence of length same as number of gates"
                )
            bounds_normals = -np.eye(self.num_inputs)

        self.bounds_normals = np.asarray(bounds_normals)
        self.bounds_limits = np.asarray(bounds_limits)

    @abstractmethod
    def compute_transition_equations(self, state_list, state):
        """
        For a given state n and a list of other states (n_1,...n_N), computes the set of
        linear equations E(v,n)-E(v,n_i). Must be implemented by derived classes.

        Parameters
        ----------
        state: K np.array of int
            State n
        state_list: NxK np.array of int
            list of other states (n_1,...n_N)

        Returns
        -------
        (A,b) set of linear equations represented by matrix A and offset b. The ith element computes
        E(v,n)-E(v,n_i) as a function of v.
        """
        pass

    @abstractmethod
    def slice(self, P, m):
        """Restricts the model to the affine subspace v=m+Pv'

        Computes the slice through the device by setting v=m+Pv', where v is the plunger gate voltages of the
        original device and v' is the new coordinate system. Must be implemented by derived classes. Note
        that derived classes also need to apply the affine transformation to the bounds variables

        Parameters
        ----------
        P: MxK np.array of floats
            The linear coefficient matrix. Here M is the number of voltage elements in v in the full simulation
            and K the dimensionality of the subspace.
        m: M np.array of floats
            offset of the affine trnsformation.

        Returns
        -------
        A model object describing the simulation on the affine subspace.
        """
        pass

    def enumerate_neighbours(self, state):
        """Generates a state_list object for compute_transition_equations,

        Enumerates the set of neighbours of a transition to return all possible
        state transitions. In this class it is implemented by returning all possible
        states that can be reached by any combintion of adding or removing an electron on any dot.
        As a result, this list has 3^D-1 elements, where D is the number of dots.

        Derived classes may overwrite this if they want to consider a more restricted set of transitions.

        Parameters
        ----------
        state: D np.array of int
            The state for which to enumerate all neighbours

        Returns
        -------
        List of lists of neighbours. By default only a list including a list of all neighbours is returned.
        Derived classes may decide to instead return several lists each representing a batch of transitions.
        Aong those lists, states do not need to be unique but instead represent groups of elements that can
        filtered efficiently by the solver, e.g., by adding all states that can be used to filter a lot of other
        states in the list quickly. This is only ever relevant when larger arrays need to be computed.
        """
        d = state.shape[0]
        # Compute list of possible state transitions for the provided state
        # For simplicity, we restrict to only single electron additions/subtractions per dot
        # This leaves 3^d-1 states
        state_list = np.zeros((1, d), dtype=int)
        for i in range(d):
            state_list1 = state_list.copy()
            state_listm1 = state_list.copy()
            state_listm1[:, i] = -1
            state_list1[:, i] = 1
            state_list = np.vstack([state_list, state_list1])
            if state[i] >= 1:
                state_list = np.vstack([state_list, state_listm1])

        # First element is all-zeros, we don't want it
        state_list = state_list[1:]

        return [state_list + state]

    def compute_polytope_for_state(self, state, maximum_slack):
        r"""For a given state, computes P(N)

        Calls enumerate_neighbours and compute_transition_equations to obtain a list of
        possible transitions and then removes from this list iteratively
        all transitions that are not sufficiently close to the polytope.

        This is computed by computing the slack. The slack is 0 if the ith transition is a facet of
        the polytope, otherwise it is a positive number computed as

        :math:`s_i = min_v A_i^Tv +b_i, v \in P(n)`

        This function retains all facets that have slack smaller than maximum_slack.
        Since enregy differences are measure din eV, the slack represents the minimum
        energy difference between the ground state and the state represented by the transition
        for any point inside the polytope.

        Parameters
        ----------
        state: D np.array of int
            The electron configuration n for which to compute P(n)
        maximum_slack: float
            The maximum distance in eV after which the transition is discarded

        Returns
        -------
        A Polytope object representing P(n). See documentation of Polytope.
        """
        # get the potentially bacthed list of states
        state_lists = self.enumerate_neighbours(state)
        if len(state_lists)>1:
            raise NotImplementedError("batching of States not implemented, yet.")
        state_list = state_lists[0]
        A, b = self.compute_transition_equations(state_list, state)

        # check, whether there are superfluous transitions
        # TODO: Oswin: i don't remember what the significance of this was.
        zero_const = np.all(np.abs(A) < 1.0e-8, axis=1)
        if np.any(zero_const):
            A = A[~zero_const]
            b = b[~zero_const]
            state_list = state_list[~zero_const]
        # ... and check for this batch whether we can filter out non-touching ones
        slacks = compute_polytope_slacks(
            A, b, self.bounds_normals, self.bounds_limits, maximum_slack
        )
        keep = slacks <= maximum_slack + 1.0e-8

        # if we have kept nothing, this means there is a set of equations that is not fullfillable
        # this happens often when slicing, e.g, a polytope is not within the sliced subspace.
        if not np.any(keep):
            return Polytope(state)

        A = A[keep]
        b = b[keep]
        slacks = slacks[keep]
        state_list = state_list[keep]

        

        # create final polytope
        poly = Polytope(state)
        touching = slacks < 1.0e-8
        point_inside, _ = compute_maximum_inscribed_circle(
            A[touching], b[touching], self.bounds_normals, self.bounds_limits
        )
        poly.set_polytope(
            state_list - state, A, b, slacks, point_inside
        )
        return poly

    def verify_polytope(self, polytope, maximum_slack):
        """Verifies a polytope.

        After slicing, polytopes that have been computed earlier also need to be sliced. It is inefficient to recompute
        the polytopes from scratch, as slicing can only remove, but never add transitions. verify_polytope allows to take
        a polytope that has been modified via polytope.layz_slice and verify/filter all transitions. This recomputes
        all slack variables and removes all transitions that have slack larger than maximum slack

        Note that this does not touch any other internal information stored in the polytope.
        This function doe snothing if polytope.must_verify=False

        TOOD: this should be moved somewhere else.

        Parameters
        ----------
        polytope: Polytope
            The polytope P(n) to be verified
        maximum_slack: float
            The maximum distance in eV after which a transition of the polytope is discarded

        Returns
        -------
        The updated polytope after filtering out transitions.
        """
        if not polytope.must_verify:
            return polytope
        slacks = compute_polytope_slacks(
            polytope.A,
            polytope.b,
            self.bounds_normals,
            self.bounds_limits,
            maximum_slack,
        )
        keep = slacks <= maximum_slack + 1.0e-8
        touching = slacks <= 1.0e-6
        point_inside, _ = compute_maximum_inscribed_circle(
            polytope.A[touching],
            polytope.b[touching],
            self.bounds_normals,
            self.bounds_limits,
        )

        verified_polytope = Polytope(polytope.state)
        verified_polytope.set_polytope(
            polytope.labels[keep],
            polytope.A[keep],
            polytope.b[keep],
            slacks[keep],
            point_inside,
        )
        return verified_polytope


class CapacitanceModel(AbstractCapacitanceModel):
    """Implementation of a slight generalization of the constant interaction model.

    The constant interaction model defines
    :math:`E(v,n)=1/2 n^T C_{DD}^{-1}n - n^T  C_{DD}^{-1}C_{DG}v`

    where :math:`C_{DD}` and  :math:`C_{DG}` are part of the maxwell capacitance matrix created by the system
    of plunger gates G and quantum dots D. Thus, :math:`C_{DD}` are the interdot capacitances (mawell) and
    :math:`C_{DG}` the dot to gate capacitances.

    This model is a generalization of the constant interaction model as it makes :math:`C_{DD}` and  :math:`C_{DG}` a function
    of electron state n. The speed of this change from the constant interaction is governed by a parameter k for each dot. The larger
    k is, the smaller the deviation. if k=None, this is exactly the constant interaction model.
    """

    def __init__(
        self,
        C_g,
        C_D,
        bounds_limits,
        bounds_normals=None,
        ks=None,
        transform_C_g=None,
        offset=None,
    ):
        """Initializes the model

        The parameters here are normal capacitances and not maxwell capacitances given in atto Farrad.

        Parameters
        ----------
        C_g: DxK np.array of float
            Capacitances in atto Farrad between the K gates and D dots.
        C_D: DxD np.array of float
            Capacitances in atto Farrad between the D dots. Self capacitances are possible via the diagonal matrix elements.
        bounds_limits: N np.array of float
            right hand side of the bound inequalities. If bounds_normals is none, this is interpreted as lower bounds and thus N=num_inputs
        bounds_normals: N x num_inputs np.array of float
            set of linear inequalities A that provide bounds for the voltage space. A valid voltage v fulfils Av+b<0, where b is
            bounds_limits. Can be none, in which case A=-Id.'
        ks: D np.array of float or None.
            How quickly the capacitances change as deviation from the constant interaction model for each dot. Can be None in which this is just the constant interaction model.
            Larger integers give smaller changes. Realistic values are 3-5.
        transform_C_g:
            Internal. Used to implement slicing. Should be None.
        offset:
            Internal. Used to implement slicing. Should be None.
        """
        # Set the transformation matrix to the identity if not provided
        if transform_C_g is None:
            transform_C_g = np.eye(C_g.shape[1])

        super().__init__(
            C_D.shape[0], transform_C_g.shape[1], bounds_limits, bounds_normals
        )

        # Set instance properties
        self.C_g_atto = np.asarray(C_g)
        self.C_D_atto = np.asarray(C_D)
        self.transform_C_g = np.array(transform_C_g)

        # Check that an offset is provided for every gate
        self.offset = np.zeros(self.transform_C_g.shape[0])
        if offset is not None:
            if len(offset) != self.transform_C_g.shape[0]:
                raise ValueError(
                    "The offset you provided does not have an offset for every gate of the device (prior to slicing)."
                )
            self.offset = np.array(offset)

        # Convert units from attoFarrad to Farrad per eV
        self.C_g = self.C_g_atto * to_Farrad_per_eV
        self.C_D = self.C_D_atto * to_Farrad_per_eV

        # Check if value for non-constant capacitance is provided
        self.ks = ks
        if ks is not None:
            self.ks = np.array(ks)
            # if np.any(ks<1):
            #    raise ValueError("The ks values must be larger than 1")

            # TODO: What are S values?
            # Cache S values for up to 1000 total dots
            self.S_cache = np.zeros((1000, self.num_dots))
            self.S_cache[0, :] = 1
            self.S_cache[1, :] = 1

            r = 2.6
            alphas = 1 - 0.137 * (1 + r) / (ks + r)
            for n in range(2, self.S_cache.shape[0]):
                Sprev = self.S_cache[n - 2]
                self.S_cache[n] = n / (
                    2 * alphas * (ks + 2) / (n + ks) + (n - 2) / Sprev
                )

    def _compute_capacitances(self, state):
        N = len(state)

        S = np.eye(N)
        if self.ks is not None:
            S_values = self.S_cache[state, range(N)]
            S = np.diag(S_values)

        sum_C_g = np.sum(self.C_g, axis=1)

        # General transform by changing dot capacitances
        Cn_g = S @ self.C_g
        Cn_D = S @ self.C_D @ S
        Csum = S @ S @ sum_C_g + np.sum(Cn_D, axis=1) + np.diag(Cn_D)

        Cn_inv = np.linalg.inv(np.diag(Csum) - Cn_D)
        return Cn_inv, Cn_g

    def compute_transition_equations(self, state_list, state_from):
        # Get number of targets
        N = state_list.shape[0]

        # Compute normal and offset for the from_state
        C0_inv, C0_g = self._compute_capacitances(state_from)
        q0 = state_from @ C0_inv
        A0 = q0 @ C0_g
        b0 = q0 @ state_from - 2 * q0 @ C0_g @ self.offset

        # Now compute the normals and offsets for the target states
        A = np.zeros((N, self.num_inputs))
        b = np.zeros(N)
        for i, n in enumerate(state_list):
            # Compute the capacitances for the target state
            Cn_inv, Cn_g = self._compute_capacitances(n)
            qn = n @ Cn_inv
            An = qn @ Cn_g

            # Compute the normal
            A[i] = (An - A0) @ self.transform_C_g
            # Compute the offset
            b[i] = (b0 - qn @ n) / 2 + qn @ Cn_g @ self.offset

        return A, b

    def slice(self, P, m):
        new_offset = self.offset + self.transform_C_g @ m
        new_transform = self.transform_C_g @ P

        new_boundsA = self.bounds_normals @ P
        new_boundsb = self.bounds_limits + self.bounds_normals @ m

        # throw out almost orthogonal bounds
        sel = np.linalg.norm(new_boundsA, axis=1) > 1.0e-7 * np.linalg.norm(
            self.bounds_normals, axis=1
        )
        new_boundsA = new_boundsA[sel]
        new_boundsb = new_boundsb[sel]

        return CapacitanceModel(
            self.C_g_atto,
            self.C_D_atto,
            new_boundsb,
            new_boundsA,
            self.ks,
            new_transform,
            new_offset,
        )



import numpy as np
from qdarts.util_functions import is_invertible_matrix
from abc import ABCMeta, abstractmethod


class AbstractPolytopeSimulator(metaclass=ABCMeta):
    """Base class for all simulation objects that can compute and return polytopes.

    The class only has a single method boundaries which returns the boundary description of the polytope
    with meta information as well as two attributes:

    Attributes
    ----------

    num_dots: int
        number of dots in the device, i.e., number of entries in the state vector of the polytope
    num_inputs: int
        number of gate voltages in the device. The polytope lives in a space that is num_inputs dimensonal.
    """

    def __init__(self, num_dots, num_inputs):
        self.num_dots = num_dots
        self.num_inputs = num_inputs

    @abstractmethod
    def boundaries(self, state):
        """
        Returns the polytope P(n) of a given state n with all its boundaries, labels and meta information.

        Parameters
        ----------
        state: list of ints
            The state n for which to compute the polytope P(n)

        Returns
        -------
        The polytope P(n)
        """
        pass

    @abstractmethod
    def slice(self, P, m, proxy=False):
        """Restricts the simulator to the affine subspace v=m+Pv'

        Computes the slice through the device by setting v=m+Pv', where v is the plunger gate voltages of the
        original device and v' is the new coordinate system. Must be implemented by derived classes.

        Parameters
        ----------
        P : MxK np.array of floats
            The linear coefficient matrix. Here M is the number of voltage elements in v in the full simulation
            and K the dimensionality of the subspace.
        m: offset of the affine trnsformation.
        proxy: bool
            Whether a proxy is returned. A proxy can share computation between instances, if supported by the derived class

        Returns
        -------
        A simulator object describing the simulation on the affine subspace. The current simulation object remains unchanged.
        """
        pass


class AbstractCapacitiveDeviceSimulator(AbstractPolytopeSimulator):
    """Base class for all objects that create device simulations from a Capacitive Model.

    This class includes all tools to compute and cache polytopes from the provided capacitive model.
    Polytopes are queried using a call to boundaries() which queries the internal cache and then
    computes the polytope on demand.

    The computed polytope P(n) is the set of voltages v for which n is the ground state of the capacitance
    energy function E(v,n), i.e., n=min_n' E(v,n'). A facet of the polytope is given by the equality
    E(v,n')-E(v,n) = 0 for a suitable choice of n'. As a result, when shooting a ray through
    a facet of the polytope created by state n', there is a transition from state n->n' in the ground state.

    The computation of the polytope discards by default all states and inequalitis that do not form a facet of P(n).
    However, this can be relaed by allowing a maximum slack, which also keeps facets for which
    min_v E(v,n')-E(v,n) < max_slack, where v is restricted to elements in the polytope P(n). i.e., max slack keeps facet
    in which the energy difference is small.

    This class supports slicing of the voltage space into affine subspaces.
    """

    def __init__(self, capacitance_model):
        super().__init__(capacitance_model.num_dots, capacitance_model.num_inputs)
        self.capacitance_model = capacitance_model
        self.cache = {}

    @abstractmethod
    def slice(self, P, m, proxy=False):
        """Restricts the simulator to the affine subspace v=m+Pv'

        Computes the slice through the device by setting v=m+Pv', where v is the plunger gate voltages of the
        original device and v' is the new coordinate system. Must be implemented by derived classes.

        Parameters
        ----------
        P : MxK np.array of floats
            The linear coefficient matrix. Here M is the number of voltage elements in v in the full simulation
            and K the dimensionality of the subspace.
        m: offset of the affine trnsformation.
        proxy: bool
            Whether a proxy is returned. A proxy shares the cache, if possible. This is the case when P is invertible,
            especially this entails M=K. If cache sharing is possible, the simulation computes the original polytope and then
            applies the affine transformation. This can reduce run time a lot if several slices need to be computed for the
            same simulation.

        Returns
        -------
        A simulator object describing the simulation on the affine subspace. The current simulation object remains unchanged.
        """
        pass

    @abstractmethod
    def compute_polytope(self, state):
        """
        Computes the polytope for a given state.

        Is implemented by the derived class and called when the polytope for a state is not found in cache.

        Parameters
        ----------
        state : list of ints
            the state identifying the polytope


        Returns
        -------
        A Polytope object containing the full computed polytope.
        """
        pass

    def compute_transition_equations(self, state_list, state_from):
        """Computes the energy difference equations from target states to all states in the list.

        For a given state and list of neighbour states, computes the linear equations Av+b that compute the energy differences
        Between the target state_from and the other states. That is, if state_list contains a list of states n', this
        function constains linear equations E(v,n')-E(v,state_from)

        Parameters
        ----------
        state_list: numpy array of ints of size NxK
            A list containing N states for which to compute the energy differences

        Returns
        -------
        A: NxK np.array, containing the linear factors for each state in state_list
        b: np.array, containing the N offsets, one for each equation.
        """
        return self.capacitance_model.compute_transition_equations(
            state_list, state_from
        )

    @abstractmethod
    def get_maximum_polytope_slack(self):
        """Returns the maximum slack value for inclusing of a facet into the polytope.

        Returns the maximum energy distance the closest point of a transition can have to the polytope
        before it is discarded. Setting to 0 means that only transitions that actually touch the polytope
        are kept.
        """
        pass

    @abstractmethod
    def set_maximum_polytope_slack(self, maximum_slack):
        """Sets the maximum slack value for inclusing of a facet into the polytope.

        Sets the maximum distance the closest point of a transition can have to the polytope
        before it is discarded. Setting to 0 means that only transitions that actually touch the polytope
        are kept.

        Note that changing this value clears the cache.
        """
        pass

    def cached_polytopes(self):
        """
        Returns a sequence including all computed and cached polytopes for inspection and modification.
        """
        return self.cache.values()

    def boundaries(self, state):
        """
        Returns the polytope P(n) of a given state n with all its boundaries, labels and meta information.

        If the polytope is not cached, it needs to be computed. This can take some time for large devices.

        Parameters
        ----------
        state: list of ints
            The state n for which to compute the polytope P(n)

        Returns
        -------
        The polytope P(n)
        """
        # Convert to array to be sure
        state = np.asarray(state).astype(int)

        # lookup key of this state
        dict_key = tuple(state.tolist())
        # See if we already have this key in our prepared list
        if dict_key not in self.cache.keys():
            self.cache[dict_key] = self.compute_polytope(state)

        # obtain polyope from dict
        polytope = self.cache[dict_key]

        # slice is allowed to be lazy but then we need to verify the polytope now.
        if polytope.must_verify:
            polytope = self.capacitance_model.verify_polytope(
                polytope, self.get_maximum_polytope_slack()
            )
            self.cache[dict_key] = polytope

        return polytope

    def inside_state(self, v, state):
        """Returns true if a point v is fully within the polytope of a given state.

        Parameters
        ----------
        state: list of ints
            The state n identifying the polytope P(n)
        v: np.array of floats
            The point v


        Returns
        -------
        The truth value of whether v is element of P(n)
        """
        polytope = self.boundaries(state)
        if len(polytope.labels) == 0:
            return False
        f = polytope.A @ v + polytope.b
        return np.all(f < 1.0e-8)

    def find_boundary_intersection(
        self,
        old_v: np.ndarray,
        new_v: np.ndarray,
        state: np.ndarray,
        epsilon: float = 1.0e-6,
        deep_search: bool = True,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Computes an intersection of a ray with the boundary of a polytope and computes the new state

        For a given state and a voltage old_v within the polytope of this state and a point new_v outside the polytope,
        computes the intersection of the ray old_v+t*(new_v-old_v) with the boundary of the polytope.
        the intersection point and new target state is computed.


        Parameters
        ----------
        old_v: np.array of floats
            A point within the current polytope
        new_v: np.array of floats
            Another point on the ray
        state: list of ints
            The ground state n of old_v. It is assumed that v is element of P(n)
        epsilon: float
            slack value added to the ray t to ensure that the point is numerically clearly outside the polytope.
        deep_search: bool
            whether an iterative search is performed for the new point in case none of the direct neighbours of the polytope match.
            If false, will throw an exception in that case. An exception is also raised when the deep search failed.


        Returns
        -------
        The first intersection point of the ray with the polytope, together with the new state
        """
        if not self.inside_state(old_v, state):
            raise ValueError("old_v must be in the provided state.")

        polytope = self.boundaries(state)

        direction = new_v - old_v
        direction /= np.linalg.norm(direction)
        selection = np.where(polytope.slacks<1.e-8)[0]
        

        A_line = polytope.A[selection] @ direction
        b_line = polytope.b[selection] + polytope.A[selection] @ old_v
        positive = np.where(A_line > 0)[0]
        ts = -b_line[positive] / A_line[positive]
        transition_idx = np.argmin(ts)
        selection = selection[positive]

        # construct point of closest hit
        transition_state = state + polytope.labels[selection[transition_idx]]
        v_intersect = old_v + (1 + epsilon) * ts[transition_idx] * direction
        if self.inside_state(v_intersect, transition_state):
            return transition_state, v_intersect

        # the new point might have went through a corner, so we check all states whose transitions are now violated

        rel_energy = polytope.A @ v_intersect + polytope.b
        idx_order = np.argsort(rel_energy)
        for idx in idx_order:
            # pass 1: ignore transitions that don't touch the polytope.
            if polytope.slacks[idx] > 1e-6:
                continue
            if rel_energy[idx] < -1.0e-8:
                continue
            transition_state = state + polytope.labels[idx]
            if self.inside_state(v_intersect, transition_state):
                return transition_state, v_intersect

        for idx in idx_order:
            # pass 2: now try the near-hits
            if polytope.slacks[idx] < 1e-6:
                continue
            if rel_energy[idx] < -1.0e-8:
                continue
            transition_state = state + polytope.labels[idx]
            if self.inside_state(v_intersect, transition_state):
                return transition_state, v_intersect
        if not self.inside_state(v_intersect, transition_state):
            if not (deep_search):
                print(old_v, new_v, state)
                raise LookupError()

            transition_state = self.find_state_of_voltage(
                new_v, state, deep_search=False
            )

        return transition_state, v_intersect

    def find_state_of_voltage(self, v, state_hint, deep_search=True):
        """Searches the ground state for a given voltage, given an initial guess.

        For a given state voltage, computes the state for which is within the polytope of the state.
        Note that the choice of the hint is not arbitrary, since the search starts from a point in state_hint
        in order to find iteratively intersections with the boundary that are closer to v. A specific requirement
        is that the polytope must not be empty, i.e., in case of a sliced simulator, the polytope must intersect
        with the affine space. This can sometimes be tricky and we recommend perform this type of computations
        only on spaces where all plungers are available and then perform the slicing through v.

        Parameters
        ----------
        v: np.array of floats
            Voltage vector for which to find the ground state
        state_hint: list of ints
            a likely candidate for the state.
        deep_search: bool
            whether an iterative search is performed in case none of the direct neighbours of a polytope match.
            If false, will throw an exception in that case. An exception is also raised when the deep search failed.
        """
        state = state_hint
        polytope = self.boundaries(state)
        if len(polytope.labels) == 0:
            raise ValueError("polytope of state_hint does not intersect with plane")

        # Check if hint was correct
        # If not we have to search.
        # We hope that the solution is close enough and find the transitions
        v_inside = polytope.point_inside.copy()
        while not self.inside_state(v, state):
            state, v_inside = self.find_boundary_intersection(
                v_inside, v, state, deep_search=deep_search
            )

        return state


class CapacitiveDeviceSimulator(AbstractCapacitiveDeviceSimulator):
    """
    This class simulates a quantum dot device based on a capacitance model.

    The simulator interally keeps track of the Coulomb diamonds (polytopes) and their transitions (facets),
    and takes care of keeping track of which transitions are feasible, with what precision, etc.
    This allows one to ask questions such as: "which transition does this facet correspond to?" and
    "what is the orthogonal axis in voltage space (i.e. virtual gate) that tunes across it?".
    The simulator will return, for each transition, a point on the transition line and the virtual gate.

    It also has the ability to take 2D slices through high dimensional voltage spaces to construct 2D
    projections of charge stability diagrams. See documentation of AbstractCapacitiveDeviceSimulator for more details.
    """

    def __init__(self, capacitance_model):
        super().__init__(capacitance_model)
        self.maximum_slack = 0.0

    def compute_polytope(self, state):
        return self.capacitance_model.compute_polytope_for_state(
            state, self.maximum_slack
        )

    def get_maximum_polytope_slack(self):
        return self.maximum_slack

    def set_maximum_polytope_slack(self, maximum_slack):
        self.maximum_slack = maximum_slack
        self.cache = {}

    def slice(self, P, m, proxy=None):
        # if proxy is not set, we check whether P is invertible
        # if it is invertible, then reusing the cache is the most efficient
        # in the general case where we don't know whether the original simulator
        # will be used still.

        # checking invertibility also allows us to quickly transform the cache
        is_invertible = is_invertible_matrix(P)
        if proxy is None:
            proxy = is_invertible

        if proxy:
            sliced_proxy = CapacitiveDeviceSimulatorProxy(self, P, m)
            return sliced_proxy
        else:
            sliced_simulator = CapacitiveDeviceSimulator(
                self.capacitance_model.slice(P, m)
            )
            sliced_simulator.maximum_slack = self.maximum_slack
            # slice all precomputed polytopes in a lazy manner.
            for key, polytope in self.cache.items():
                if is_invertible:
                    sliced_simulator.cache[key] = polytope.invertible_transform(P, m)
                else:
                    sliced_simulator.cache[key] = polytope.lazy_slice(P, m)

            return sliced_simulator


class CapacitiveDeviceSimulatorProxy(AbstractCapacitiveDeviceSimulator):
    """
    This class is a slice proxy for the CapacitiveDeviceSimulator class. It gets returned by
    any slice operation, when a proxy is requested. This is unlikely to be used by the user
    directly and mostly used during plotting. The advantage of a proxy is that it can make better use of
    caching at the expense of higher computation cost: all queries for polytopes are computed by the original simulator
    and thus if several different slices of the same simulator are needed, they can share computed polytopes.

    For the methods, see the documentation of AbstractCapacitiveDeviceSimulator
    """

    def __init__(self, simulator, P, m):
        super().__init__(simulator.capacitance_model.slice(P, m))
        self.simulator = simulator
        self.P = P
        self.m = m

    def get_maximum_polytope_slack(self):
        return self.simulator.get_maximum_polytope_slack()

    def set_maximum_polytope_slack(self, maximum_slack):
        self.simulator.set_maximum_polytope_slack(maximum_slack)
        self.cache = {}

    def compute_polytope(self, state):
        # query or compute original polytope
        polytope = self.simulator.boundaries(state)

        # transform lazyly

        polytope_sliced = polytope.lazy_slice(self.P, self.m)
        polytope_sliced = self.capacitance_model.verify_polytope(
            polytope_sliced, self.get_maximum_polytope_slack()
        )
        return polytope_sliced

    def slice(self, P, m, proxy=None):
        if proxy is None:
            proxy = True

        if proxy:
            return CapacitiveDeviceSimulatorProxy(self, P, m)
        else:
            new_P = self.P @ P
            new_m = self.m + self.P @ m
            return self.simulator.slice(new_P, new_m, False)



from qdarts.simulator import CapacitiveDeviceSimulator
from qdarts.tunneling_simulator import (
    TunnelBarrierModel,
    ApproximateTunnelingSimulator,
    NoisySensorDot,
)
from qdarts.noise_processes import OU_process, Cosine_Mean_Function
from qdarts.capacitance_model import CapacitanceModel
from qdarts.util_functions import (
    find_point_on_transitions,
    fix_gates,
    find_label,
    compensate_simulator_sensors,
    axis_align_transitions,
)

# for the algorithm
import numpy as np

# for plotting
from matplotlib import pyplot as plt
from qdarts.plotting import get_CSD_data, get_polytopes

# for measuring runtime

plt.rcParams.update({"font.size": 18})

# SIMULATION CONSTANTS
SLOW_NOISE = {"tc": 50}


class Experiment:  # TODO: change name to the simulator name
    """Experiment class that helps setup the different parts of the simulator into an easy to ue interface.
    The usage of this class is best described by the tutorials in the example folder.
    """

    def __init__(
        self,
        capacitance_config,
        tunneling_config=None,
        sensor_config=None,
        print_logs=True,
    ):
        """
        capacitance_config: dictionary containing the capacitance model parameters
        tunneling_config: dictionary containing the tunneling model parameters
        sensor_config: dictionary containing the sensor model parameters
        print_logs: bool, whether to print logs
        """
        # Store configurations
        self.capacitance_config = capacitance_config
        self.tunneling_config = tunneling_config
        self.sensor_config = sensor_config
        self.print_logs = print_logs
        
        if print_logs:
            print("EXPERIMENT INITIALIZED")
            print("-----------------------")

        # Deploy simulators
        self.capacitance_sim = self.deploy_capacitance_sim(capacitance_config)
        self.has_sensors = False

        # Check requirements for sensor and tunneling configurations
        if tunneling_config is not None and sensor_config is None:
            raise ValueError(
                "Specifying a tunneling configuration also requires a sensor configuration."
            )
        if tunneling_config is None and sensor_config is not None:
            raise ValueError(
                "Specifying a sensor configuration also requires a tunneling configuration."
            )
        if tunneling_config is not None:
            self.sensor_model = self.deploy_sensor_model(sensor_config)
            self.tunneling_sim = self.deploy_tunneling_sim(
                self.capacitance_sim, tunneling_config
            )
            self.has_sensors = True

    # DEPLOYMENT FUNCTIONS
    # --------------------------
    def deploy_capacitance_sim(self, config):
        """
        Function that deploys a capacitance simulator.

        Arguments
        ---------
        config: dictionary containing the capacitance model parameters

        Returns
        -------
        sim: CapacitiveDeviceSimulator object
        """
        capacitance_model = CapacitanceModel(
            config["C_Dg"], config["C_DD"], -1, ks=config["ks"]
        )
        sim = CapacitiveDeviceSimulator(capacitance_model)

        # Save the parameters
        self.N = len(config["C_Dg"])  # number of dots
        self.inner_dots = list(
            np.arange(self.N)
        )  # indices of the dots. NOTE: no sensor at this point

        if self.print_logs:
            if config["ks"] is not None:
                # Print log of capacitance parameters
                log = """
                Capacitance model deployed with the following parameters:
                Dot-dot capacitances: 
                {}
                Dot-gate capacitances: 
                {}
                Size of Coulomb peaks V[n] scale as {}/(n + {})
                """.format(
                    config["C_DD"],
                    config["C_Dg"],
                    np.round(
                        (1 - 0.137 * 3.6 / (config["ks"] + 2.6) * config["ks"]), 3
                    ),
                    2 + config["ks"]
                )
                print(log)
            else:
                # Print log of capacitance parameters
                log = """
                Capacitance model deployed with the following parameters:
                Dot-dot capacitances: 
                {}
                Dot-gate capacitances: 
                {}
                Size of Coulomb peaks V[n] is constant
                """.format(config["C_DD"], config["C_Dg"])
                print(log)

        return sim

    def deploy_tunneling_sim(self, capacitance_sim, tunneling_config):
        """
        Function that deploys a tunneling simulator.

        Arguments
        ---------
        capacitance_sim: CapacitiveDeviceSimulator object
        tunneling_config: dictionary containing the tunneling model parameters

        Returns
        -------
        tunneling_sim: ApproximateTunnelingSimulator object
        """
        tunneling_matrix = tunneling_config["tunnel_couplings"]
        tunnel_barrier_offsets = np.log(tunneling_matrix + 1e-20)
        barrier_levers = (
            tunneling_config["barrier_levers"]
            if "barrier_levers" in tunneling_config.keys()
            else None
        )
        barrier_sim = TunnelBarrierModel(tunnel_barrier_offsets, barrier_levers)
        tunneling_sim = ApproximateTunnelingSimulator(
            capacitance_sim,
            barrier_sim,
            tunneling_config["temperature"],
            self.sensor_model,
        )

        # explore the neighboring polytopes
        slack = tunneling_config["energy_range_factor"] * np.maximum(
            1.0 / tunneling_sim.beta, np.max(tunneling_matrix)
        )
        tunneling_sim.poly_sim.set_maximum_polytope_slack(slack)
        tunneling_sim.num_additional_neighbours[self.sensor_model.sensor_dot_ids] = 2

        if self.print_logs:
            # Print log of tunneling parameters
            log = """
            Tunneling model deployed with the following parameters:
            Tunneling matrix:
            {}
            Temperature: {} K
            Energy range factor: {}
            """.format(
                tunneling_matrix,
                tunneling_config["temperature"],
                tunneling_config["energy_range_factor"],
            )
            print(log)

        return tunneling_sim

    def deploy_sensor_model(self, sensor_config):
        """
        Function that deploys a sensor model.

        Arguments
        ---------
        config: dictionary containing the sensor model parameters

        Returns
        -------
        sensor_sim: NoisySensorDot object
        """
        if "noise_amplitude" not in sensor_config.keys():
            sensor_config["noise_amplitude"] = {}
        if "fast_noise" not in sensor_config["noise_amplitude"]:
            sensor_config["noise_amplitude"]["fast_noise"] = 0.0
        if "slow_noise" not in sensor_config["noise_amplitude"]:
            sensor_config["noise_amplitude"]["slow_noise"] = 0.0
        if "signal_noise" not in sensor_config["noise_amplitude"]:
            sensor_config["noise_amplitude"]["signal_noise"] = 0.0
        # Separate between inner and sensor dots
        self.inner_dots = list(
            set(self.inner_dots) - set(sensor_config["sensor_dot_indices"])
        )

        # Define slow-noise generator
        slow_noise_gen = OU_process(  # TODO: LATER: implement 1/f noise
            sigma=sensor_config["noise_amplitude"]["slow_noise"],
            tc=SLOW_NOISE["tc"],
            dt=1,  # the unit of time is the single measurment
            num_elements=len(sensor_config["sensor_dot_indices"]),
        )

        if "mean_field" in sensor_config.keys():
            mean_field_config = sensor_config["mean_field"]
            if mean_field_config["type"] == "Cosine_Mean_Function":
                b = mean_field_config["b"] if "b" in mean_field_config.keys() else None
                slow_noise_gen = Cosine_Mean_Function(
                    slow_noise_gen, mean_field_config["a"], mean_field_config["W"], b
                )

        # Deploy sensor model
        sensor_sim = NoisySensorDot(sensor_config["sensor_dot_indices"])

        # Configure sensor model
        sensor_sim.config_noise(
            sigma=sensor_config["noise_amplitude"]["fast_noise"],
            signal_noise_scale=sensor_config["noise_amplitude"]["signal_noise"],
            slow_noise_gen=slow_noise_gen,
        )
        sensor_sim.config_peak(
            g_max=1.0, peak_width_multiplier=sensor_config["peak_width_multiplier"]
        )

        # Pring log of sensor parameters
        if self.print_logs:
            log = """
            Sensor model deployed with the following parameters:   
            Sensor dot indices: {}
            Sensor detunings: {} meV
            Coulomb peak width: {} meV
            Slow noise amplitude: {} ueV
            Fast noise amplitude: {} ueV
            Signal noise scale: {}
            """.format(
                sensor_config["sensor_dot_indices"],
                np.array(sensor_config["sensor_detunings"]) * 1e3,
                np.round(
                    2
                    * sensor_config["peak_width_multiplier"]
                    * self.tunneling_config["temperature"]
                    * 86
                    * 1e-3
                    / 0.631,
                    2,
                ),
                np.round(sensor_config["noise_amplitude"]["slow_noise"] * 1e6, 4),
                np.round(sensor_config["noise_amplitude"]["fast_noise"] * 1e6, 4),
                np.round(sensor_config["noise_amplitude"]["signal_noise"], 5),
            )
            print(log)
        return sensor_sim

    def center_transition(
        self,
        simulator,
        target_state,
        target_transition,
        plane_axes,
        use_virtual_gates=False,
        compensate_sensors=False,
    ):
        """
        Function that center the CSD at a given facet (transition) of the polytope (occupation state).

        Arguments
        ---------
        simulator: any simulator object
        target_state: int, the state at which the transition happens, e.g. [2,2]
        target_transition: list of integers, the transition point e.g. [1,-1] would be the transition from [2,2] to [1,1]
        plane_axes: 2xN array, the axes of the transition which span the plane
        use_virtual_gates: bool, whether to use virtual gates
        compensate_sensors: bool, whether to compensate the sensors

        Returns
        -------
        plane_axes: 2xN array, the axes spanning the cut through voltage plane
        transition_sim: CapacitanceSimulator object, the transition simulator
        """
        if compensate_sensors:
            # fix the sensor gate if we compensate
            simulator = fix_gates(
                simulator,
                self.sensor_config["sensor_dot_indices"],
                np.zeros(len(self.sensor_config["sensor_dot_indices"])),
            )
            # reduce dimension of the voltage space
            complement_gates = [
                i
                for i in range(plane_axes.shape[1])
                if i not in self.sensor_config["sensor_dot_indices"]
            ]
            plane_axes = plane_axes[:, complement_gates]
        # find boundaries of the selected polytope
        poly = simulator.boundaries(target_state)
        # get_labels of the transitions
        inner_labels = poly.labels[:, self.inner_dots]

        # Find the index of the transition point
        transition = np.array(target_transition)[self.inner_dots].T
        try:
            idx_multidot_transition = [find_label(inner_labels, transition)[0]]
        except ValueError:
            raise "The transition point is not in the polytope. Please choose a different transition"

        # Find the offset of V
        v_transition = find_point_on_transitions(poly, idx_multidot_transition)
        print("v_offset found:", v_transition)

        if use_virtual_gates:
            # Compute the normals
            pair_transitions = np.array(
                [np.array(transition).T for transition in plane_axes], dtype=int
            )
            idxs = [
                find_label(inner_labels, t[self.inner_dots])[0]
                for t in pair_transitions
            ]
            normals = -poly.A[idxs]
            normals /= np.linalg.norm(normals, axis=1)[:, None]
            P_transition = normals.T
            transition_sim = axis_align_transitions(
                simulator.slice(P_transition, v_transition),
                target_state,
                poly.labels[idxs],
                [0, 1],
            )
        else:
            transition_sim = simulator.slice(plane_axes.T, v_transition)

        return np.eye(2), transition_sim

    # GETTERS
    # --------------------------

    def get_virtualised_sim(self, simulator, target_state):
        """
        Function that takes a simulator and virtualises the gates specified by inner_dots.

        Arguments
        ---------
        simulator: AbstractPolytopeSimulator object
        target_state: int, the initial corner state guess

        Returns
        -------
        simulator: AbstractPolytopeSimulator object, the virtualised simulator
        """
        gate_transitions = np.eye(self.N, dtype=int)[self.inner_dots]
        # TODO: Default target state is lower left corner state (initial guess). In future user could specify!
        self.target_state = target_state
        simulator = axis_align_transitions(
            simulator, self.target_state, gate_transitions, self.inner_dots
        )
        return simulator

    def get_compensated_sim(self, simulator, target_state):
        """
        Function that takes a capacitance simulator and compensates the sensors.

        Arguments
        ---------
        simulator: AbstractPolytopeSimulator object
        target_stater: int, the state at which sensor compensation happens

        Returns
        -------
        simulator: AbstractPolytopeSimulator object, the compensated simulator
        """
        if not self.has_sensors:
            raise ValueError("Compensating sensors requires a sensor model.")

        sim,v = compensate_simulator_sensors(
            simulator,
            target_state=target_state,
            compensation_gates=self.sensor_config["sensor_dot_indices"],
            sensor_ids=self.sensor_config["sensor_dot_indices"],
            sensor_detunings=self.sensor_config["sensor_detunings"]
        )
        
        #translate sensor gates to 0.
        offset = np.zeros(len(v))
        offset[self.sensor_config["sensor_dot_indices"]] = v[self.sensor_config["sensor_dot_indices"]]
        return sim.slice(np.eye(len(v)),offset)
        

    def get_plot_args(self, x_voltages, y_voltages, plane_axes, v_offset=None):
        """
        Function that returns the arguments for plotting the CSD.

        Arguments
        ---------
        x_voltages: list of floats, the x-axis voltages
        y_voltages: list of floats, the y-axis voltages
        plane_axes: 2xN array, the axes of the plane in which the CSD is to be rendered
        v_offset: Nx1 array, the offset voltage of all of the gates, which defines the origin of the plot

        Returns
        -------
        v_offset: Nx1 array, the offset voltage of all of the gates
        minV: 2x1 array, the minimum voltage of selected axes
        maxV: 2x1 array, the maximum voltage of selected axes
        resolution: list of integers, the resolution of the plot
        """

        if v_offset is None:
            v_offset = np.zeros(plane_axes.shape[1], dtype=float)
        else:
            v_offset = np.array(v_offset, dtype=float)

        xout = x_voltages + np.dot(v_offset, plane_axes[0])
        yout = y_voltages + np.dot(v_offset, plane_axes[1])

        minV = np.array([x_voltages[0], y_voltages[0]])
        maxV = np.array([x_voltages[-1], y_voltages[-1]])

        resolution = [len(x_voltages), len(y_voltages)]

        return v_offset, minV, maxV, resolution, xout, yout

    # RENDER FUNCTIONS
    # --------------------------

    def generate_CSD(
        self,
        x_voltages,
        y_voltages,
        plane_axes,
        target_state=None,
        target_transition=None,
        use_virtual_gates=False,
        compensate_sensors=False,
        compute_polytopes=False,
        use_sensor_signal=False,
        v_offset=None,
        insitu_axis=None,
    ):
        """
        Function that renders the capacitance CSD for a given set of voltages and axes.

        Arguments
        ---------
        x_voltages: list of floats, the x-axis voltages
        y_voltages: list of floats, the y-axis voltages
        plane_axes: 2xN array, the axes of the plane in which the CSD is to be rendered
        target_state: int, the guess state or the state at which the transition happens
        target_transition: list of integers, the transition point e.g. [1,-1] would be the transition from [2,2] to [1,1]
        use_virtual_gates: bool, whether to use virtual gates
        compensate_sensors: bool, whether to compensate the sensors
        compute_polytopes: bool, whether to compute the polytopes
        use_sensor_signal: bool, whether to use the sensor signal
        v_offset: Nx1 array, the offset voltage of all of the gates, which defines the origin of the plot

        Returns
        -------
        xout, yout: list of floats, the x and y voltages
        CSD_data: 2D array, the CSD data
        polytopes: dictionary, the polytopes of the CSD. None if compute_polytopes is False
        sensor_values: 3D array, the sensor signal [size(xout),size(yout),num_sensors]. None if use_sensor_signal is False
        v_offset: Nx1 array, the offset voltage of all of the gates
        """
        # check required parameters
        if target_state is None:
            # if not target state use [0,0,0,0,0,0]
            target_state = [0] * self.N

        plane_axes = np.array(plane_axes)
        # prepare plot
        v_offset, minV, maxV, resolution, xout, yout = self.get_plot_args(
            x_voltages, y_voltages, plane_axes, v_offset
        )

        # pick the simulator
        if use_sensor_signal:
            simulator = self.tunneling_sim
        else:
            simulator = self.capacitance_sim
        if compensate_sensors:
            simulator = self.get_compensated_sim(simulator, target_state=target_state)
        if target_transition is not None:
            plane_axes, simulator = self.center_transition(
                simulator,
                target_state,
                target_transition,
                plane_axes,
                use_virtual_gates,
                compensate_sensors,
            )
            v_offset = np.zeros(2)  # TODO: how to do it nicer?

        elif use_virtual_gates:
            simulator = self.get_virtualised_sim(simulator, target_state)

        if use_sensor_signal:
            csimulator = simulator.poly_sim
        else:
            csimulator = simulator

        # Part for the electrostatic CSD:
        CSD_data = None
        polytopes = None
        if not use_sensor_signal or compute_polytopes:
            backend, CSD_data, states = get_CSD_data(
                csimulator,
                np.array(plane_axes).T,
                v_offset,
                minV,
                maxV,
                resolution,
                target_state,
            )
        if compute_polytopes:
            polytopes = get_polytopes(states, backend, minV, maxV)

        # Part for the sensor signal:
        sensor_values = None
        if use_sensor_signal:
            # TODO: How does this work ? The long chain of condition makes it hard
            # to understand what's going on ... Arguments like plane_axes share the
            # same name but end up having a completely different shape depending on the flags
            sensor_values = simulator.sensor_scan_2D(
                plane_axes,
                v_offset,
                minV,
                maxV,
                resolution,
                target_state,
                insitu_axis=insitu_axis,
            )

        return xout, yout, CSD_data, polytopes, sensor_values, v_offset



import numpy as np
from abc import ABCMeta, abstractmethod


class AbstractNoiseProcess(metaclass=ABCMeta):
    r"""
    Models a noise process that produces possibly dependent samples :math:`\epsilon(v)_t`. The noise process
    can depend on the device voltages v and the noise can depend on all previous samples in the sequence. We assume
    that :math:`\epsilon(v)_t` is vector valued and the number of elements is stored in the ``num_elements`` attribute

    A sequence is started by calling ``start_sequence``, at which point the newly sampled points are independent from
    all previous samples.

    Note that currently, the elements in the noise process are assumed to be drawn with equal time difference between samples.

    Attributes
    ----------
        num_elements: int
            the dimensionality of the noise variable
    """

    def __init__(self, num_elements):
        self.num_elements = num_elements

    @abstractmethod
    def slice(self, P, m):
        """Restricts the noise to the affine subspace :math:`v=m+Pv`.

        Parameters
        ----------
        P : MxK np.array of floats
            The linear coefficient matrix.
        m: M np.array of floats
            offset of the affine transformation.

        Returns
        -------
        A noise process object describing the noise on the affine subspace. The current noise object remains unchanged.
        """
        pass

    @abstractmethod
    def start_sequence(self):
        """
        Restart the sequence so that the next sample drawn is independent from all previous samples.
        """
        pass

    @abstractmethod
    def __call__(self, v):
        r"""Returns the next element of the noise process.

        Parameters
        ----------
        v : K np.array of floats
            The current voltage parameter of the device

        Returns
        -------
        An element of the noise process :math:`\epsilon(v)_t`. This is assumed to be a vector of M elements,
        """
        pass


class OU_process(AbstractNoiseProcess):
    r"""Implements the Ornstein-Uhlenbeck noise process

    This noise is independent of v. We have that

    .. math::
        \epsilon(v)_{t+1}=a \epsilon(v)_t + \sqrt{1-a^2} \sigma \epsilon_t

    Where :math:`\epsilon_t` is independent gaussian noise and

    .. math::
        a=e^{-\\frac{\Delta t}{t_c}}

    It is possible to generate multiple independent samples of the same process simultaneously.
    """

    def __init__(self, sigma, tc, dt, num_elements):
        """Instantiates the OU process with its process parameters.

        Arguments
        ---------
        sigma: float
            Standard deviation of the OU process
        tc: float
            time correlation parameters, the higher the more samples are correlated
        dt: float
            time step between samples. The higher, the less samples are correlated
        num_elements: int
            How many independnet samples are computed simultaneously
        """
        super().__init__(num_elements)
        self.sigma = sigma
        self.tc = tc
        self.dt = dt
        self.a = np.exp(-self.dt / self.tc)
        self.b = np.sqrt(1 - np.exp(-2 * self.dt / self.tc)) * self.sigma

    def slice(self, P, m):
        return self

    def start_sequence(self):
        self.x = self.sigma * np.random.randn(self.num_elements)

    def next_val(self):
        n = np.random.randn(len(self.x))
        self.x = self.x * self.a + self.b * n
        return self.x

    def __call__(self, v):
        return self.next_val()


class Cosine_Mean_Function(AbstractNoiseProcess):
    r"""Decorator of an random process that models an additive mean term that depends on the gate voltages.
    This term is added to noise values sampled from the decorated noise model

    The mean term of the ith noise element is given as a set of cosine functions:

    :math:`\mu_i(v)= \sum_j a_{ij} \cos(2\pi (w_{ij}^T v+b_{ij}))`

    and the resulting noise is

    :math:`\epsilon(v)_{t}=\mu(v) + \epsilon(v)^D_{t}`

    where :math:`\epsilon(v)^D_{t}` is the decorated noise process.

    The user supplies the weight tensor W with elements :math:`W_{ijk}` so that W[i,j] is the vector :math:`w_{ij}`
    and a matrix a with the amplitude values :math:`a_{ij}`. Finally, b is the matrix of offsets :math:`b_{ij}`, which can be left as None,
    in which case it is sampled uniformly between 0 and 1.
    """

    def __init__(self, noise_model, a, W, b=None):
        """Initialized the cosine mean function.

        Parameters
        ----------
        noise_model: AbstractNoiseProcess
            The noise process with N dimensions to decorrate.
        a: NxM np.array of float
            Amplitudes of the M overlapping cosine functions
        W: NxMxK np.array of float
            K dimensional weights of the M consine functions for the N outputs.
        b: NxM np.array of float or None
            Phases of the M overlapping cosine functions. If none, it is drawn uniformly between 0 and 1.
        """
        super().__init__(noise_model.num_elements)
        self.noise_model = noise_model
        self.a = a
        self.W = W
        self.b = b if b is not None else np.random.uniform(size=a.shape)

    def slice(self, P, m):
        new_W = np.einsum("ijk,kl->ijl", self.W, P)
        new_b = self.b + np.einsum("ijk,k->ij", self.W, m)
        return Cosine_Mean_Function(self.noise_model, self.a, new_W, new_b)

    def start_sequence(self):
        self.noise_model.start_sequence()

    def __call__(self, v):
        # compute noise
        noise_values = self.noise_model(v)

        # add the mean
        activation = 2 * np.pi * (np.einsum("ijk,k->ij", self.W, v) + self.b)
        mean = np.sum(2 * np.pi * self.a * np.cos(activation), axis=1)
        noise_values += mean.reshape(1, -1)

        return noise_values



import cvxpy as cp
import scipy
import clarabel
import numpy as np
from scipy.spatial import HalfspaceIntersection


def is_sequence(seq):
    if isinstance(seq, str):
        return False
    try:
        len(seq)
    except Exception:
        return False
    return True


def is_invertible_matrix(A, max_cond=1.0e8):
    """Returns true if A is an invertible matrix.

    Parameters
    ----------
    max_cond: float
        conditioning of A (fraction of maximum and minimum absolute eigenvalue) above which
        it is assumed that A is numerically rank-deficient.
    """
    if A.shape[0] != A.shape[1]:
        return False
    else:
        return np.linalg.cond(A) < max_cond


def solve_linear_problem(prob):
    """Internal helper function to solve supplied linear cvxpy problems"""
    try:
        prob.solve(verbose=False, solver=cp.CLARABEL, max_iter=1000000)
    except cp.SolverError:
        prob.solve(solver=cp.GLPK)

def solve_linear_ineq_problem(cx,bx, A_ineq,b_ineq):
    """Internal helper function to solve LP with no equality constraints"""

    # convert to SCS format
    N = len(cx)
    M = A_ineq.shape[0]
    P = scipy.sparse.csc_matrix((N,N))
    q = -cx
    A = scipy.sparse.csc_matrix(A_ineq)
    b = -b_ineq
    
    cone = [clarabel.NonnegativeConeT(M)]

    # Set solver parameters
    settings = clarabel.DefaultSettings()
    settings.verbose = False
    
    #initializeand solve
    solver = clarabel.DefaultSolver(P,q,A,b,cone,settings)
    solution = solver.solve()
    
    status = str(solution.status)
    success = False
    if status == "Solved" or status == "AlmostSolved":
        success = True
    
    if success :
        return True, -solution.obj_val+bx, solution.x
    else:
        return False, None, None
    
def _compute_polytope_slacks_1D(A, b):
    """Special case called by compute_polytope_slacks when A has a single column"""
    w = A.reshape(-1)
    close_to_zero = np.abs(w) < 1.0e-6 * np.max(w)
    # compute the single point fulfilling the constraint
    x = -b / (w + close_to_zero * 1.0e-6 * np.max(w))

    # check whether the constraint is a lower or upper bound
    is_lower = w < 0
    # count the constraints for checking whether we have both lower and upper bounds
    num_lower = np.sum(is_lower)

    # now find the lower and upper bounds of the interval that fulfill all constraints
    if num_lower > 0:
        lower_x = np.max(x[is_lower])
    else:
        lower_x = -np.inf
    if num_lower < len(A):
        upper_x = np.min(x[~is_lower])
    else:
        upper_x = np.inf
    # handling of infeasibility
    # since the slacks are defined as smallest violation of the constraints,
    # if the polytope is infeasible, the slack for the lower bound constraints
    # must be computed based on the upper bound and vice versa
    if lower_x > upper_x:
        temp = lower_x
        lower_x = upper_x
        upper_x = temp

    # compute slacks:
    slacks_lower = -(is_lower * (w * lower_x + b))
    slacks_upper = -((~is_lower) * (w * upper_x + b))

    slacks = slacks_lower + slacks_upper
    return slacks


def _compute_polytope_slacks_2D(A, b, bounds_A, bounds_b):
    """Special case in 2D solved via halfspace intersection"""
    
    # first we use halfspace intersection to compute all corners of the final polytope
    # find a point fulfilling all constraints
    feasible_point, _ = compute_maximum_inscribed_circle(A, b, bounds_A, bounds_b)
    # bring all constraints together in matrix form
    halfspaces_poly = np.concatenate([A, b.reshape(-1, 1)], axis=1)
    halfspaces_bounds = np.concatenate([bounds_A, bounds_b.reshape(-1, 1)], axis=1)
    halfspaces = np.concatenate([halfspaces_poly, halfspaces_bounds], axis=0)
    hs = HalfspaceIntersection(halfspaces, feasible_point)
    intersections = hs.intersections

    # now compute slacks. since the slack is computed as the solution of a linear programming problem,
    # the solution must lie on one of the vertices.
    slacks_intersections = A @ (intersections.T) + b.reshape(-1, 1)

    min_slacks = np.min(-slacks_intersections, axis=1)
    slacks = np.maximum(min_slacks, np.zeros(min_slacks.shape))

    return slacks


def compute_polytope_slacks(A, b, bounds_A, bounds_b, maximum_slack):
    """Computes the slacks of each candidate transition of a ground state polytope.

    The polytope is given by all points x, such that :math:`Ax+b<0`. There might be boundaries
    such that for no x holds that :math:`A_ix+b_i = 0`. In this case the definition of the polytope is
    the same when it is removed. However, sometimes we are interested in keeping transitions that
    are near misses - i.e., there exists an x such that, the inequality is almost fulfilled.
    In this case, we can relax this by allowing a positive slack and
    accept transitions to still be relevant for the polytope when we find an x, such that

    :math:`A_ix+b_i <= s`

    if this inequality holds exactly with slack :math:`s=0`, we say that the ith transition touches
    the polytope and the larger slack is, the more distant is the polytope

    Computing the slack can be difficult in the presence of unbounded polytopes. For this reason additional linear bound
    constraints need to be provided that ensure that all polytopes are bounded.

    The function computes the minimum slack for all transitions in :math:`A_i` and :math:`b_i`.
    The default slack is needed if for some reason it is not possible to compute the slack due to numerical
    difficulties.

    Parameters
    ----------
    A: NxK np.array of floats
        The linear coefficients of the N affine linear equations in K dimensions
    b: N np.array of floats
        The constant offsets of the N affine linear equations
    bounds_A: N'xK np.array of floats
        The linear coefficients of N' additional constraint that ensure that polytopes are bounded.
    bounds_b: N' np.array of floats
        The constant offsets of the N' bounds
    maximum_slack: float
        Value for the maximum acceptable slack for transitions to be considered near the polytope.
        TODO: review whether this parameter is needed.
    """

    # to find whether an equation a^Tx+b<= 0 touches the polytope, we need to find
    # a point that is on the intersection between polytope and the line where
    # a^Tx+b=0. For this we systematiclly go throuh the full list of
    # equations and for each solve an LP with this goal. if an equation
    # is found not to touch the polytope it is removed from the list
    # of candidates for all future solves.

    # we slightly generalize this problem by weakening the equation via
    # a slack variable
    # a^Tx+b = eps
    # eps > 0
    # and minimize for eps.
    # if eps = 0, then the equation touches the polytope. If not, then
    # it can be removed. this allows us to introduce a maximum slack value
    # which allows us to keep close matches for later stages in the simulation, e.g., to keep
    # neighbours that are possibly relevant for tunneling equations. Note: for eps>0 this formulation
    # produces the x value in the polytope closest to the equation in function value a^Tx+b.

    # we return the vector of eps values for all equations so that the user can filter transitions afterwards

    # first special cases.
    # only one constraint? feasible qwith slack 0
    if len(b) == 1:
        return np.zeros(1)

    # 1D and 2D problems can be solved efficiently
    if A.shape[1] == 1:
        return _compute_polytope_slacks_1D(A, b)
    if A.shape[1] == 2:
        return _compute_polytope_slacks_2D(A, b, bounds_A, bounds_b)

    # now we know there is a polyope and we can compute its sides
    N = len(A)
    touching = np.ones(
        N, dtype=bool
    )  # equations with eps~=0. At the beginning we assume all are touching
    slacks = (maximum_slack + 1) * np.ones(
        N
    )  # slack value (updated when equation is computed)
    A_touch = A
    b_touch = b
    for k in range(N):
        success, value, x = solve_linear_ineq_problem(A[k],b[k], A_touch,b_touch)
        touching[k]=False
        if success:
            slacks[k] = np.maximum(-value,0.0)
            if slacks[k] < 1.0e-6:
                touching[k] = True
        if not touching[k]:
            A_touch = A[touching, :]
            b_touch = b[touching]
    return slacks


def compute_maximum_inscribed_circle(A, b, bounds_A, bounds_b):
    """Computes the maximum inscribed circle in a polytope intersected with a set of linear inequalities.

    The maximum inscribed circle is a crude measure for position and size of a polytope.
    It computes the circle with maximum radius r and midpoint m, such that all its points
    lie inside the polytope. The function returns the (m,r) maximizing this. This choice is very often not unique.

    Since the polytope given by linear equations A,b might be unbounded, the function takes another
    set of linear equations for establishing lower and upper bounds. In essence, this is the same as
    adding the additional equalities to A and b and computing the maximum inscribed circle for that polytope.



    Parameters
    ----------
    A: NxK np.array of floats
        The linear coefficients of the N affine linear equations in the K-dimensional polytope
    b: N np.array of floats
        The constant offsets of the N affine linear equations of the polytope
    bounds_A: MxK np.array of floats
        The linear coefficients of the M added linear inequality constraints
    bounds_b: M np.array of floats
        The constant offsets of the M added linear inequality constraints
    """
    if len(A) == 0:
        return None, 0.0

    norm_A = np.linalg.norm(A, axis=1)
    norm_bounds = np.linalg.norm(bounds_A, axis=1)

    r = cp.Variable()
    v = cp.Variable(A.shape[1])
    constraints = [
        A @ v + b + r * norm_A
        <= 0,  # linear boundaries are only allowed to intersect the sphere once
        bounds_A @ v + bounds_b + norm_bounds * r
        <= 0,  # also stay away from bound constraints
        r >= 0,  # Radius is strictly positive
    ]
    prob = cp.Problem(cp.Maximize(r), constraints)
    solve_linear_problem(prob)
    return v.value, r.value


def find_label(labels, label):
    """helper function that finds the position of a state in a matrix of states"""
    dist = np.sum(np.abs(labels - np.array(label)), axis=1)
    return np.where(dist < 1.0e-5)[0]


def find_point_on_transitions(polytope, indizes):
    """Finds a point on a facet (or intersection point of multiple facets) of a polytope

    Given a precomputed polytope with facets (A,b) and their slacks,
    computes a point where a set of facet inequalities are exactly equal.
    Among all points that fulfill this, we pick the mid point defined
    by the maximum inscribed circle on the facet (or the subfacet
    created by the intersection of facets).

    TODO: it is not quite clear what happens when the indizes are not touching.

    Parameters
    ----------

    polytope: A polytope object
        The polytope for which the intersections are computed
    indices: list of int
        The subset of facets for which a common point is to be found.
        note that it is quietly assumed that the facets are touching the polytope.
    """
    slacks = np.delete(polytope.slacks, indizes)
    A = np.delete(polytope.A, indizes, axis=0)
    b = np.delete(polytope.b, indizes)
    A = A[slacks < 1.0e-8]
    b = b[slacks < 1.0e-8]
    A_eq = polytope.A[indizes, :]

    b_eq = polytope.b[indizes]

    norms = np.linalg.norm(A, axis=1)

    eps = cp.Variable()
    x = cp.Variable(A.shape[1])
    prob = cp.Problem(
        cp.Maximize(eps), [A @ x + b + norms * eps <= 0, A_eq @ x + b_eq == 0, eps >= 0]
    )
    solve_linear_problem(prob)
    return x.value


def fix_gates(simulator, gate_ids, gate_values, proxy=False):
    """Fixes a number of gate voltages in the simulator object

    Returns a new simulation where the values of the given gates are fixed to a constant.
    This is done by computing the apropriate parameters for slice, and therefore the
    operation can not be undone in the returned simulation.

    Please keep in mind that by doing this, all indices of gate voltages at entries
    after the deleted entries change, i.e., in a device with 4 plungers, removing the
    third plungers will lead to a simulator with 3 plungers where the last plunger has
    index 3. It is therefore advisable to order parameters such that fix_gates is always
    applied to the end.

    Parameters
    ----------
    simulator: AbstractPolytopeSimulator
        The simulator object for which gates are to be fixed
    gate_ids: list of int
        The indices of the gates in the voltage vector
    gate_values: np.array of float
        The values of the fixed gates
    proxy: bool
        whether or not the returned simulator should be a proxy, i.e., share cache if possible
        Todo: is this ever possible?
    """
    v = np.zeros(simulator.num_inputs)
    v[gate_ids] = gate_values

    P = np.zeros((simulator.num_inputs, simulator.num_inputs - len(gate_ids)))
    pos = 0
    for i in range(simulator.num_inputs):
        if i not in gate_ids:
            P[i, pos] = 1
            pos += 1
    return simulator.slice(P, v, proxy)


def axis_align_transitions(
    simulator, target_state, transitions, compensation_gates, proxy=True
):
    """Transform the simulators coordinate system such that transitions are aligned with coordinate axes

    Takes a set of transitions from a target state and a set of gate indices of same length.
    Computes a linear transformation such, that the normal of the ith transition is parallel to the ith gate axis supplied as
    argument.

    For example, to align the transition from state [1,1,1] to [1,1,2] with the first plunger gate, we set
    target_state=[1,1,1], transitions=[[1,0,0]] and compensation_gates=[0]

    Parameters
    ----------
    simulator: AbstractPolytopeSimulator
        The simulator object which is to be transformed
    target_state: list of int
        The state from which the transitions are extracted
    transitions: NxD np.array of int
        The set of N transitions (given as D-dimensional difference vectors state-target_state) to align
    compensation_gates: list of int
        The indices of the plunger gates that should be transformed to align with the transition normals
    proxy: bool
        whether or not the returned simulator should be a proxy, i.e., share cache if possible
    """
    compensation_gates = np.array(compensation_gates, dtype=int)
    # get the polytope of the target state
    polytope = simulator.boundaries(target_state)

    # find the transitions inside the polytope
    transition_idxs = []
    for transition in transitions:
        idx = find_label(polytope.labels, transition)[0]
        transition_idxs.append(idx)

    # get normals of the transitions
    normals = -polytope.A[transition_idxs, :]
    # normalize to ensure that we do not change the spacing of transitions
    normals /= np.linalg.norm(normals, axis=1)[:, None]

    # compute compensation matrix
    B = normals[:, compensation_gates]
    compensation = -B.T @ np.linalg.inv(B @ B.T)

    # compute coordinate transform
    P = np.eye(simulator.num_inputs)

    # get the indizes of the elements in the submatrix of the compensation parameters
    P_sub_ids = (
        simulator.num_inputs * compensation_gates[:, None] + compensation_gates[None, :]
    )
    np.put(P, P_sub_ids, compensation.flatten())

    return simulator.slice(P, np.zeros(simulator.num_inputs), proxy=proxy)

def compute_sensor_compensation_transform(
    simulator,
    target_state,
    compensation_gates,
    sensor_ids,
    sensor_detunings,
    sensor_slope_detuning=0.0,
):
    """Transforms the simulation to compensate the sensors against all other gates.

    This function allows for perfect or imperfect sensor compensation as well as the exact position on the sensor peak.
    This is done by finding the compensation values of the sensor plunger gates to compensate for the linear cross-talk of all other
    plungers. This compensation is computed for a given target state as the compensation parameters might depend on the capacitances
    in the state if they are variable.

    The position on the sensor peak is given by sensor_detunings which move the position as a direct modification of the sensor potential.

    Parameters
    ----------
    simulator: AbstractPolytopeSimulator
        The simulator object which is to be transformed
    target_state: list of int
        The state from which the transitions are extracted
    compensation_gates: list of int
        The gates to be used for compensation of the sensor. Typically the sensor plunger gates in the device
    sensor_ids: list of int
        The indices of the sensor ids.
    sensor_detunings: np.array of float
        detuning parameter for each sensor which allows to move the sensor on a pre-specified point of the peak.
    sensor_slope_detuning: float
        (Experimental) scaling factor that moves the compensation linearly from perfect compensation (0) to no compensation (1).

    Returns
    -------
    P: the compensation matrix
    compute_bias(v): computes the offset b so that P@v+b is a point with the given detuning
    v: a point with detuning = 0, the point the compensation is computed about
    """
    if len(sensor_ids) != len(compensation_gates):
        raise ValueError(
            "Number of gates for compensation must equal number of sensors"
        )

    if len(sensor_ids) != len(sensor_detunings):
        raise ValueError(
            "Number of gates for compensation must equal number of sensors"
        )

    for sensor in sensor_ids:
        if target_state[sensor] <= 0:
            raise ValueError(
                "Target state must have at least one electron on each sensor dot"
            )

    compensation_gates = np.array(compensation_gates, dtype=int)
    other_gates = np.delete(np.arange(simulator.num_inputs), compensation_gates)
    sensor_detunings = np.array(sensor_detunings)

    # by default we assume that for the sensor dots,
    # we compute the transition between dots K and K+1
    # where K is the electron occupation on the target state,
    # which means that we take the polytope at the target state
    # and compute the transition for the electron K->K+1 on the sensor dot.
    # However, we will change the computed polytope based
    # on the sensor detuning. if it is positive, we will instead compute
    # the polytope for the K+1 electron and then search for the transition K+1->K

    target_state = target_state.copy()
    transitions = []
    detunings = []
    for detuning, sens_id in zip(sensor_detunings, sensor_ids):
        if detuning > 0:
            transitions.append(-np.eye(1, simulator.num_dots, sens_id))
            detunings.append(-detuning)
        else:
            target_state[sens_id] -= 1
            transitions.append(np.eye(1, simulator.num_dots, sens_id))
            detunings.append(detuning)

    # get geometry of the target state to compensate for
    polytope = simulator.boundaries(target_state)

    # find the sensor transitions inside the polytope
    transition_idxs = []
    for transition in transitions:
        idx = find_label(polytope.labels, transition)[0]
        transition_idxs.append(idx)

    # get normals of sensor transitions
    normals = polytope.A[transition_idxs, :]
    # compute point on the intersection of the transition
    v = find_point_on_transitions(polytope, transition_idxs)
    

    # compute compensation matrix
    #normals /= np.linalg.norm(normals, axis=1)[:, None]
    A1 = normals[:, compensation_gates]
    A2 = normals[:, other_gates]
    compensation = -np.linalg.inv(A1) @ A2
    # now create the P-matrix
    P = np.eye(simulator.num_inputs)
    # get the indizes of the elements in the submatrix of the compensation parameters
    P_sub_ids = (
        simulator.num_inputs * compensation_gates[:, None] + other_gates[None, :]
    )
    np.put(P, P_sub_ids, compensation.flatten())

    # add errors to P-matrix
    P_error = (1 - sensor_slope_detuning) * P + sensor_slope_detuning * np.eye(
        simulator.num_inputs
    )
    
    #offset function. 
    
    # apply sensor detunings. First compute virtual gate matrix P_S for the sensor voltages
    P_S = A1.T @ np.linalg.inv(A1 @ A1.T)
    v_detuned = v.copy()
    v_detuned[compensation_gates] += P_S@detunings
    offset_v = v_detuned - P[:, other_gates] @ v[other_gates]
    
    
    def compute_offset(v):
        #using only the non-sensor gates, compute the sensor point with given detuning
        v_comp = P[:, other_gates] @ v[other_gates] + offset_v
        #b is the offset so that P_error@v+b=v_comp
        b = v_comp-P_error@v
        return b
    
    return P_error,  compute_offset, v_detuned

def compensate_simulator_sensors(
    simulator,
    target_state,
    compensation_gates,
    sensor_ids,
    sensor_detunings,
    sensor_slope_detuning=0.0,
):
    """Transforms the simulation to compensate the sensors against all other gates.

    This function allows for perfect or imperfect sensor compensation as well as the exact position on the sensor peak.
    This is done by finding the compensation values of the sensor plunger gates to compensate for the linear cross-talk of all other
    plungers. This compensation is computed for a given target state as the compensation parameters might depend on the capacitances
    in the state if they are variable.

    The position on the sensor peak is given by sensor_detunings which move the position as a direct modification of the sensor potential.

    Parameters
    ----------
    simulator: AbstractPolytopeSimulator
        The simulator object which is to be transformed
    target_state: list of int
        The state from which the transitions are extracted
    compensation_gates: list of int
        The gates to be used for compensation of the sensor. Typically the sensor plunger gates in the device
    sensor_ids: list of int
        The indices of the sensor ids.
    sensor_detunings: np.array of float
        detuning parameter for each sensor which allows to move the sensor on a pre-specified point of the peak.
    sensor_slope_detuning: float
        (Experimental) scaling factor that moves the compensation linearly from perfect compensation (0) to no compensation (1).

    Returns
    -------
    sliced_sim: the sliced simulation that is created from the computed compensation parameters
    tuning_point: a vector of gate voltages that indicates the exact compensation point of the simulation
    """
    
    P_error, compute_bias, v = compute_sensor_compensation_transform(
        simulator,
        target_state,
        compensation_gates,
        sensor_ids,
        sensor_detunings,
        sensor_slope_detuning
    )
    offset = compute_bias(v)
    
    return simulator.slice(P_error, offset, True), v


import numpy as np
from scipy.spatial import HalfspaceIntersection, ConvexHull
from scipy.optimize import linprog
from qdarts.util_functions import is_sequence
from tqdm import tqdm


def find_feasible_point(halfspaces):
    """Computes a feasible point by a polytope defined in halfspace format. internal."""
    norm_vector = np.reshape(
        np.linalg.norm(halfspaces[:, :-1], axis=1), (halfspaces.shape[0], 1)
    )
    c = np.zeros((halfspaces.shape[1],))
    c[-1] = -1
    A = np.hstack((halfspaces[:, :-1], norm_vector))
    b = -halfspaces[:, -1:]
    res = linprog(c, A_ub=A, b_ub=b, bounds=(None, None))
    return res.x[:-1]


def plot_2D_polytope(
    ax, A, b, color, lower_bounds, label=None, linestyle="-", linewidth=1
):
    """Plots a single 2D polytope. internal."""
    eqs = np.hstack([A, b.reshape(-1, 1)])
    # add constraints to polytope
    eqs = np.vstack([eqs, lower_bounds])

    # Get the corners

    feasible_point = find_feasible_point(eqs)
    if not np.all(eqs[:, :-1] @ feasible_point + eqs[:, -1] < 1.0e-5):
        return None, None
    corners = HalfspaceIntersection(eqs, feasible_point).intersections

    # Easiest way to get proper line-segments of the boundaries in 2D is via the convex hull
    # Form a convex hull based on the corners
    hull = ConvexHull(corners)
    # Plot lines between the points forming the hull (they are assumed to come in counterclockwise order)
    for i, simplex in enumerate(hull.simplices):
        label_trace = label
        if i != 0:
            label_trace = None
        ax.plot(
            corners[simplex, 0],
            corners[simplex, 1],
            linestyle,
            c=color,
            lw=linewidth,
            label=label_trace,
        )
    return feasible_point, (np.min(corners, axis=0), np.max(corners, axis=0))


def get_2D_polytope(A, b, color, lower_bounds, label=None, linestyle="-", linewidth=1):
    """computes the corners of a 2D polytope from the provided polytope. internal"""
    eqs = np.hstack([A, b.reshape(-1, 1)])
    # add constraints to polytope
    eqs = np.vstack([eqs, lower_bounds])

    # Get the corners

    feasible_point = find_feasible_point(eqs)
    if not np.all(eqs[:, :-1] @ feasible_point + eqs[:, -1] < 1.0e-5):
        return None, None
    corners = HalfspaceIntersection(eqs, feasible_point).intersections

    # Easiest way to get proper line-segments of the boundaries in 2D is via the convex hull
    # Form a convex hull based on the corners
    return corners


def raster_CSD_states(
    simulation, P, v_0, minV, maxV, resolution, state_hint_lower_right
):
    """Creates a grid of points in 2D space and computes for each point the ground state. Internal."""
    if not is_sequence(resolution):
        resolution = [resolution, resolution]
    states = np.zeros((resolution[0], resolution[1], simulation.num_dots), dtype=int)
    line_start = simulation.find_state_of_voltage(
        v_0 + P @ np.array([minV[0], minV[1]]), state_hint_lower_right
    )
    pbar = tqdm(total=np.prod(resolution), desc="Rastering CSD")
    # TODO: this is terrible in terms of performance
    for i, v1 in enumerate(np.linspace(minV[0], maxV[0], resolution[0])):
        state = line_start
        for j, v2 in enumerate(np.linspace(minV[1], maxV[1], resolution[1])):
            v = v_0 + P @ np.array([v1, v2])
            state = simulation.find_state_of_voltage(v, state)
            if not simulation.inside_state(v, state):
                print("error")
            if j == 0:
                line_start = state
            states[i, j] = state
            pbar.update(1)
    return states


def get_CSD_data(
    simulation, P, v_0, lower_left, upper_right, resolution, state_hint_lower_left,proxy=False
):
    """
    Function that computes a Charge Stability Diagram from a simulation of a device.
    The function plots the states at voltages v=v_0+P*x where x is a vector with elements
    in a box defined by its lower-left and upper-right corners given by lower_left and upper_right and
    number of values given by the resolution. if upper_right=-lower_left then v_0 is the center
    pixel of the rastered plot.

    By default, a background image is plotted
    which is based on the charge configuration at a position. This is overlaid with a line plot indicating the
    exact transition points between states. Optionally each region is labelled using the exact electron state.

    Parameters
    ----------
        simulation:
            the device simulation to raster
        ax:
            matplotlib axis object to plot into
        v_0:
            the origin of the coordinate system to plot.
        P:
            coordinate system. A nx2 axis where n is the number of gates in simulation.
        lower_left:
            minimum value in x for both axes
        upper_right:
            maximum value of x in both axes
        resolution:
            number of sampled points in each direction of x. either single number or one per axis.
        state_hint_lower_left:
            starting point to guess the initial state of the lower left corner. must not be empty within sim.
        proxy:
            Whether to cache the computations of the full polytopes. False is faster in most cases.
    """
    minV = np.array(lower_left)
    maxV = np.array(upper_right)

    # find the true corner state before slicing the simulation. otherwise we might have trouble
    # finding it if the state hint does not touch the projection

    lower_left_corner = v_0 + P @ minV
    corner_state = simulation.find_state_of_voltage(
        lower_left_corner, state_hint_lower_left
    )
    simulation_slice = simulation.slice(P, v_0, proxy=proxy)

    # compute CSD
    states = raster_CSD_states(
        simulation_slice, np.eye(2), np.zeros(2), minV, maxV, resolution, corner_state
    )
    color_weights = np.linspace(1, 2.7, simulation_slice.num_dots)
    CSD_data = 1 + np.sum(color_weights.reshape(1, 1, -1) * states, axis=2)
    return simulation_slice, CSD_data, states


def get_polytopes(states, simulation_slice, minV, maxV):
    """For each unique state in the provided state list, computes the corners of the polytope. of the 2D sliced simulation.

    This function is used for plotting of the exact state lines of the underlying capacitive model.
    """
    # iterate over the list of different states and plot their sliced polytope

    states = [tuple(s) for s in states.reshape(-1, simulation_slice.num_dots).tolist()]
    state_list = set(states)
    polytope_list = {}
    for state in state_list:
        # get the polytope
        polytope = simulation_slice.boundaries(state)
        A = polytope.A
        b = polytope.b

        # check if polytope is empty and continue otherwise (should never trigger)
        if A.shape[0] == 0:
            continue
        lower_bounds_graph = np.hstack([-np.eye(2), (minV - 0.05 * abs(minV))[:, None]])
        upper_bounds_graph = np.hstack([np.eye(2), -(maxV + 0.05 * abs(maxV))[:, None]])
        lower_bounds = np.vstack([lower_bounds_graph, upper_bounds_graph])
        corners = get_2D_polytope(A, b, "white", lower_bounds)
        polytope_list[str(state)] = np.array(corners)
    return polytope_list


def plot_polytopes(
    ax,
    polytopes,
    fontsize=10,
    color="w",
    axes_rescale=1,
    only_labels=False,
    only_edges=False,
    skip_dots=[],
    alpha=1,
    lw=1,
):
    """Plot the polytopes computes by get_polytopes"""
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    for polytope in polytopes:
        corners = polytopes[polytope]
        hull = ConvexHull(polytopes[polytope])
        if not only_labels:
            for i, simplex in enumerate(hull.simplices):
                ax.plot(
                    corners[simplex, 0] * axes_rescale,
                    corners[simplex, 1] * axes_rescale,
                    "-",
                    c=color,
                    lw=lw,
                    alpha=alpha,
                )
        if not only_edges:
            box_mid = (
                (np.array(np.min(corners, axis=0)) + np.array(np.max(corners, axis=0)))
                / 2
                * axes_rescale
            )
            # skip skip_dots entries to polytope
            res = np.array([int(ele) for ele in str(polytope[1:-1]).split(",")])
            inds = np.array(list(set(np.arange(len(res))) - set(skip_dots)), dtype=int)

            if (
                box_mid[0] > xlim[0]
                and box_mid[0] < xlim[1]
                and box_mid[1] > ylim[0]
                and box_mid[1] < ylim[1]
            ):
                ax.text(
                    box_mid[0],
                    box_mid[1],
                    str(res[inds].tolist()),
                    c=color,
                    ha="center",
                    va="center",
                    fontsize=fontsize,
                )
